#include <algorithm>
#include <vector>

#include "caffe/filler.hpp"
#include "caffe/layers/base_conv_hash_layer.hpp"
#include "caffe/util/MyMacro.h"
#include "caffe/util/hash2col.hpp"

namespace caffe {

template <typename Dtype>
BaseConvHashLayer<Dtype>::~BaseConvHashLayer()
{
#ifdef USE_CUDNN
	cudnnDestroyTensorDescriptor(bottom_desc_);
	cudnnDestroyTensorDescriptor(top_desc_);
	cudnnDestroyConvolutionDescriptor(conv_desc_);
	cudnnDestroyFilterDescriptor(filter_desc_);
	cudnnDestroy(handle_);
#endif // USE_CUDNN
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::LayerSetUp(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) {

	//CHECK size
	if (bottom.size()!= HASH_DATA_SIZE)
	{
		printf("Fatal error: bottom size should be %d\n", HASH_DATA_SIZE);
		exit(0);
	}
	if (top.size()!= HASH_DATA_SIZE)
	{
		printf("Fatal error: top size should be %d\n", HASH_DATA_SIZE);
		exit(0);
	}

  // Configure the kernel size, padding, stride, and inputs.
  const ConvHashParameter &conv_param = this->layer_param_.conv_hash_param();
  
  num_spatial_axes_ = 3;	//for 3D convolution

  vector<int> spatial_dim_blob_shape(1, num_spatial_axes_);
  // Setup filter kernel dimensions (kernel_shape_).
  kernel_shape_.Reshape(spatial_dim_blob_shape);
  int* kernel_shape_data = kernel_shape_.mutable_cpu_data();
  {
	  const int num_kernel_dims = conv_param.kernel_size_size();
	  CHECK(num_kernel_dims == 1 || num_kernel_dims == num_spatial_axes_)
		  << "kernel_size must be specified once, or once per spatial dimension "
		  << "(kernel_size specified " << num_kernel_dims << " times; "
		  << num_spatial_axes_ << " spatial dims).";
	  for (int i = 0; i < num_spatial_axes_; ++i) {
		  kernel_shape_data[i] =
			  conv_param.kernel_size((num_kernel_dims == 1) ? 0 : i);
	  }
  }
  for (int i = 0; i < num_spatial_axes_; ++i) {
    CHECK_GT(kernel_shape_data[i], 0) << "Filter dimensions must be nonzero.";
  }
  // Setup stride dimensions (stride_).
  stride_.Reshape(spatial_dim_blob_shape);
  int* stride_data = stride_.mutable_cpu_data();
  {
    const int num_stride_dims = conv_param.stride_size();
    CHECK(num_stride_dims == 0 || num_stride_dims == 1 ||
          num_stride_dims == num_spatial_axes_)
        << "stride must be specified once, or once per spatial dimension "
        << "(stride specified " << num_stride_dims << " times; "
        << num_spatial_axes_ << " spatial dims).";
    const int kDefaultStride = 1;
    for (int i = 0; i < num_spatial_axes_; ++i) {
      stride_data[i] = (num_stride_dims == 0) ? kDefaultStride :
          conv_param.stride((num_stride_dims == 1) ? 0 : i);
      CHECK_GT(stride_data[i], 0) << "Stride dimensions must be nonzero.";
    }
  }
  // Setup pad dimensions (pad_).
  pad_.Reshape(spatial_dim_blob_shape);
  int* pad_data = pad_.mutable_cpu_data();
  {
	  const int num_pad_dims = conv_param.pad_size();
	  CHECK(num_pad_dims == 0 || num_pad_dims == 1 ||
		  num_pad_dims == num_spatial_axes_)
		  << "pad must be specified once, or once per spatial dimension "
		  << "(pad specified " << num_pad_dims << " times; "
		  << num_spatial_axes_ << " spatial dims).";
	  const int kDefaultPad = 0;
	  for (int i = 0; i < num_spatial_axes_; ++i) {
		  pad_data[i] = (num_pad_dims == 0) ? kDefaultPad :
			  conv_param.pad((num_pad_dims == 1) ? 0 : i);
	  }
  }
  // Special case: im2col is the identity for 1x1 convolution with stride 1
  // and no padding, so flag for skipping the buffer and transformation.
  is_1x1_ = true;
  for (int i = 0; i < num_spatial_axes_; ++i) {
	  is_1x1_ &=
		  kernel_shape_data[i] == 1 && stride_data[i] == 1 && pad_data[i] == 0;
	  if (!is_1x1_) { break; }
  }
  // Configure output channels and groups.
  channels_ = bottom[HASH_DATA_BLOB]->shape(CHANNEL_AXIS);
  num_output_ = conv_param.num_output();
  CHECK_GT(num_output_, 0);
  if (reverse_dimensions()) {
	  conv_out_channels_ = channels_;
	  conv_in_channels_ = num_output_;
  }
  else {
	  conv_out_channels_ = num_output_;
	  conv_in_channels_ = channels_;
  }
  // Handle the parameters: weights and biases.
  // - blobs_[0] holds the filter weights
  // - blobs_[1] holds the biases (optional)
  vector<int> weight_shape(2);
  weight_shape[0] = conv_out_channels_;
  weight_shape[1] = conv_in_channels_;
  for (int i = 0; i < num_spatial_axes_; ++i) {
	  weight_shape.push_back(kernel_shape_data[i]);
  }
  bias_term_ = conv_param.bias_term();
  vector<int> bias_shape(bias_term_, num_output_);
  if (this->blobs_.size() > 0) {
	  CHECK_EQ(1 + bias_term_, this->blobs_.size())
		  << "Incorrect number of weight blobs.";
	  if (weight_shape != this->blobs_[0]->shape()) {
		  Blob<Dtype> weight_shaped_blob(weight_shape);
		  LOG(FATAL) << "Incorrect weight shape: expected shape "
			  << weight_shaped_blob.shape_string() << "; instead, shape was "
			  << this->blobs_[0]->shape_string();
	  }
	  if (bias_term_ && bias_shape != this->blobs_[1]->shape()) {
		  Blob<Dtype> bias_shaped_blob(bias_shape);
		  LOG(FATAL) << "Incorrect bias shape: expected shape "
			  << bias_shaped_blob.shape_string() << "; instead, shape was "
			  << this->blobs_[1]->shape_string();
	  }
	  LOG(INFO) << "Skipping parameter initialization";
  }
  else {
	  if (bias_term_) {
		  this->blobs_.resize(2);
	  }
	  else {
		  this->blobs_.resize(1);
	  }
	  // Initialize and fill the weights:
	  // output channels x input channels per-group x kernel height x kernel width
	  this->blobs_[0].reset(new Blob<Dtype>(weight_shape));
	  shared_ptr<Filler<Dtype> > weight_filler(GetFiller<Dtype>(
		  conv_param.weight_filler()));
	  weight_filler->Fill(this->blobs_[0].get());
	  // If necessary, initialize and fill the biases.
	  if (bias_term_) {
		  this->blobs_[1].reset(new Blob<Dtype>(bias_shape));
		  shared_ptr<Filler<Dtype> > bias_filler(GetFiller<Dtype>(
			  conv_param.bias_filler()));
		  bias_filler->Fill(this->blobs_[1].get());
	  }
  }

  kernel_dim_ = this->blobs_[0]->count(1);
  // Propagate gradients to the parameters (as directed by backward pass).
  this->param_propagate_down_.resize(this->blobs_.size(), true);

  //check whether conv on self 
  m_bottom_struct_idx = conv_param.bottom_struct();
  m_top_struct_idx = conv_param.top_struct();
  conv_on_self_ = (m_bottom_struct_idx == m_top_struct_idx);

  if (reverse_dimensions()) {
	  m_conv_bt_struct_idx = m_top_struct_idx;
	  m_conv_top_struct_idx = m_bottom_struct_idx;
  }
  else {
	  m_conv_bt_struct_idx = m_bottom_struct_idx;
	  m_conv_top_struct_idx = m_top_struct_idx;
  }

#ifdef USE_CUDNN
  // cudnn
  CUDNN_CHECK(cudnnCreate(&handle_));
  cudnn::createFilterDesc<Dtype>(&filter_desc_, conv_out_channels_,
	  conv_in_channels_, kernel_shape_data[0]* kernel_shape_data[1]* kernel_shape_data[2], 1);
  cudnn::createTensor4dDesc<Dtype>(&bottom_desc_);
  cudnn::createTensor4dDesc<Dtype>(&top_desc_);
  cudnn::createConvolutionDesc<Dtype>(&conv_desc_);
  filter_workspace_size_ = 8 * 1024 * 1024;
  vector<int> filter_workspace_shape(1, filter_workspace_size_);
  filter_workspace_.Reshape(filter_workspace_shape);
#endif //USE_CUDNN
}


template <typename Dtype>
void BaseConvHashLayer<Dtype>::reshape_topHashData(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{
	//const ConvHashParameter &conv_param = this->layer_param_.conv_hash_param();
	const Blob<int>* top_defNumSum_blob = CGlobalSys<Dtype>::get_defNumSum_blob(m_top_struct_idx);
	const int total_top_defNum = top_defNumSum_blob->cpu_data()[top_defNumSum_blob->count() - 1];
	std::vector<int> data_shape;
#if ADD_NUM_2_SHAPE	//to consistent with N*C*SPATIAL_DIM, for the loss of segmentation
	data_shape.push_back(1);
#endif
	data_shape.push_back(num_output_);
	data_shape.push_back(total_top_defNum);
	top[HASH_DATA_BLOB]->Reshape(data_shape);
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::Reshape(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) 
{    
  //reshape the top hash data
  reshape_topHashData(bottom, top);

  //reshape the top hash data
  reshape_colBuf(bottom, top);
}

//NOTE: this function must be called after reshape top data
template <typename Dtype>
void BaseConvHashLayer<Dtype>::reshape_colBuf(const vector<Blob<Dtype>*>& bottom, const vector<Blob<Dtype>*>& top)
{
	if (reverse_dimensions()) {
		conv_out_spatial_dim_ = bottom[HASH_DATA_BLOB]->count(SPATIAL_AXIS);
	}
	else {
		conv_out_spatial_dim_ = top[HASH_DATA_BLOB]->count(SPATIAL_AXIS);
	}

	Blob<Dtype> &col_buf = CGlobalSys<Dtype>::get_workspace();

	std::vector<int> col_buf_shape;
	col_buf_shape.push_back(kernel_dim_);
	col_buf_shape.push_back(conv_out_spatial_dim_);
	col_buf.Reshape(col_buf_shape);
	
	out_spatial_dim_ = top[HASH_DATA_BLOB]->count(SPATIAL_AXIS);

	//also reshape the bias multiplier
	if (bias_term_) 
	{
		Blob<Dtype> &bias_multiplier = CGlobalSys<Dtype>::get_multiplier();

		vector<int> bias_multiplier_shape(1, out_spatial_dim_);
		bias_multiplier.Reshape(bias_multiplier_shape);


		switch (Caffe::mode())
		{
		case Caffe::CPU:
			caffe_set(bias_multiplier.count(), Dtype(1), bias_multiplier.mutable_cpu_data());
			break;
		case Caffe::GPU:
			caffe_gpu_set(bias_multiplier.count(), Dtype(1), bias_multiplier.mutable_gpu_data());
			break;
		default:
			break;
		}
	}

#if RECORD_NEIGHBOR
	const int* kernel_shape_data = kernel_shape_.cpu_data();
	std::vector<int> nei_shape;
	int kernel_spatial_dim = kernel_shape_data[0] * kernel_shape_data[1] * kernel_shape_data[2];
	nei_shape.push_back(kernel_spatial_dim);
	nei_shape.push_back(conv_out_spatial_dim_);
	nei_idx_.Reshape(nei_shape);
#endif
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::forward_cpu_gemm(const Dtype* input,
	const Dtype* weights, Dtype* output)
{
	if (conv_on_self_)
	{
		//printf("conv forward CPU on self!\n");
		forward_cpu_gemm_self(input, weights, output);
	}else
	{
		//printf("conv forward CPU with down sampling!\n");
		forward_cpu_gemm_ds(input, weights, output);
	}
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::forward_cpu_gemm_self(const Dtype* input,
	const Dtype* weights, Dtype* output)
{
	if (!conv_on_self_)
	{
		printf("Error: should be conv on self!\n");
		exit(0);
	}
	const Dtype* col_buff = input;
	if (!is_1x1_) 
	{
		CBatchStructure self_struct;
		CGlobalSys<Dtype>::setBatchStructure(self_struct, m_top_struct_idx);	//top is the same as bottom
		Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
		//hash2col_self_cpu(input, channels_, &self_struct, kernel_shape_.cpu_data(), global_buf.mutable_cpu_data());
		hash2col_self_cpu(input, conv_in_channels_, &self_struct, kernel_shape_.cpu_data(), global_buf.mutable_cpu_data());
		col_buff = global_buf.cpu_data();
	}
	caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, conv_out_channels_, conv_out_spatial_dim_, kernel_dim_,
		(Dtype)1., weights, col_buff,
		(Dtype)0., output);	
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::forward_cpu_gemm_ds(const Dtype* input,
	const Dtype* weights, Dtype* output)
{
	if (conv_on_self_)
	{
		printf("Error: should be conv with stride > 1!\n");
		exit(0);
	}
	if (is_1x1_)
	{
		printf("Unexpected Error: is_1x1_ cannot happen in stride > 1!\n");
		exit(0);
	}

	CBatchStructure bt_struct, tp_struct;
	CGlobalSys<Dtype>::setBatchStructure(bt_struct, m_conv_bt_struct_idx);
	CGlobalSys<Dtype>::setBatchStructure(tp_struct, m_conv_top_struct_idx);
	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
	hash2col_ds_cpu(input, conv_in_channels_, &bt_struct, &tp_struct, kernel_shape_.cpu_data(), stride_.cpu_data(), pad_.cpu_data(), global_buf.mutable_cpu_data());
	const Dtype *col_buff = global_buf.cpu_data();
	
	caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, conv_out_channels_, conv_out_spatial_dim_, kernel_dim_,
		(Dtype)1., weights, col_buff,
		(Dtype)0., output);
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::forward_cpu_bias(Dtype* output, const Dtype* bias)
{
	const Blob<Dtype> &bias_multiplier = CGlobalSys<Dtype>::get_multiplier();
	caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, num_output_,
	  out_spatial_dim_, 1, (Dtype)1., bias, bias_multiplier.cpu_data(),
	  (Dtype)1., output);
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::backward_cpu_gemm(const Dtype* output,
	const Dtype* weights, Dtype* input)
{
	if (conv_on_self_)
	{
		//printf("conv backward gemm CPU on self!\n");
		backward_cpu_gemm_self(output, weights, input);
	}
	else
	{
		//printf("conv backward gemm CPU with down sampling!\n");
		backward_cpu_gemm_ds(output, weights, input);
	}
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::backward_cpu_gemm_self(const Dtype* output,
	const Dtype* weights, Dtype* input)
{
	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
	Dtype* col_buff = global_buf.mutable_cpu_data();
	if (is_1x1_)
	{
		col_buff = input;
	}

	caffe_cpu_gemm<Dtype>(CblasTrans, CblasNoTrans, kernel_dim_,
		conv_out_spatial_dim_, conv_out_channels_,
		(Dtype)1., weights, output,
		(Dtype)0., col_buff);

	if (!is_1x1_)
	{
		CBatchStructure self_struct_cpu;
		CGlobalSys<Dtype>::setBatchStructure(self_struct_cpu, m_top_struct_idx);	//top is the same as bottom

		//col2hash_self_cpu(input, channels_, &self_struct_cpu, kernel_shape_.cpu_data(),
		//	col_buff);
		col2hash_self_cpu(input, conv_in_channels_, &self_struct_cpu, kernel_shape_.cpu_data(),
			col_buff);
	}
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::backward_cpu_gemm_ds(const Dtype* output,
	const Dtype* weights, Dtype* input)
{
	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
	Dtype* col_buff = global_buf.mutable_cpu_data();

	caffe_cpu_gemm<Dtype>(CblasTrans, CblasNoTrans, kernel_dim_,
		conv_out_spatial_dim_, conv_out_channels_,
		(Dtype)1., weights, output,
		(Dtype)0., col_buff);


	CBatchStructure bt_struct_cpu, top_struct_cpu;
	CGlobalSys<Dtype>::setBatchStructure(bt_struct_cpu, m_conv_bt_struct_idx);	
	CGlobalSys<Dtype>::setBatchStructure(top_struct_cpu, m_conv_top_struct_idx);

	col2hash_ds_cpu(input, conv_in_channels_, &bt_struct_cpu, &top_struct_cpu, kernel_shape_.cpu_data(), stride_.cpu_data(), pad_.cpu_data(), col_buff);
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::weight_cpu_gemm(const Dtype* input,
	const Dtype* output, Dtype* weights)
{
	if (conv_on_self_)
	{
		//printf("conv weight gemm CPU on self\n");
		weight_cpu_gemm_self(input, output, weights);
	}
	else
	{
		//printf("conv weight gemm CPU with down sampling!\n");
		weight_cpu_gemm_ds(input, output, weights);
	}
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::weight_cpu_gemm_self(const Dtype* input,
	const Dtype* output, Dtype* weights)
{
	const Dtype* col_buff = input;
	if (!is_1x1_)
	{
		CBatchStructure self_struct;
		CGlobalSys<Dtype>::setBatchStructure(self_struct, m_top_struct_idx);	//top is the same as bottom
		Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
		//hash2col_self_cpu(input, channels_, &self_struct, kernel_shape_.cpu_data(), global_buf.mutable_cpu_data());
		hash2col_self_cpu(input, conv_in_channels_, &self_struct, kernel_shape_.cpu_data(), global_buf.mutable_cpu_data());
		col_buff = global_buf.cpu_data();
	}

	caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasTrans, conv_out_channels_,
		kernel_dim_, conv_out_spatial_dim_,
		(Dtype)1., output, col_buff,
		(Dtype)0., weights);
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::weight_cpu_gemm_ds(const Dtype* input,
	const Dtype* output, Dtype* weights)
{
	CBatchStructure bt_struct, top_struct;
	CGlobalSys<Dtype>::setBatchStructure(bt_struct, m_conv_bt_struct_idx);	
	CGlobalSys<Dtype>::setBatchStructure(top_struct, m_conv_top_struct_idx);
	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
	hash2col_ds_cpu(input, conv_in_channels_, &bt_struct, &top_struct,kernel_shape_.cpu_data(), stride_.cpu_data(),pad_.cpu_data(),global_buf.mutable_cpu_data());
	const Dtype *col_buff = global_buf.cpu_data();


	caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasTrans, conv_out_channels_,
		kernel_dim_, conv_out_spatial_dim_,
		(Dtype)1., output, col_buff,
		(Dtype)0., weights);
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::backward_cpu_bias(Dtype* bias,
	const Dtype* input)
{
	Blob<Dtype> &bias_multiplier = CGlobalSys<Dtype>::get_multiplier();

	caffe_cpu_gemv<Dtype>(CblasNoTrans, num_output_, out_spatial_dim_, 1.,
		input, bias_multiplier.cpu_data(), (Dtype)0., bias);
}


#ifndef CPU_ONLY
template <typename Dtype>
void BaseConvHashLayer<Dtype>::forward_gpu_gemm(const Dtype* input,
	const Dtype* weights, Dtype* output)
{
	if (conv_on_self_)
	{
		//printf("conv forward GPU on self!\n");
		forward_gpu_gemm_self(input, weights, output);
	}
	else
	{
		//printf("conv forward GPU with down sampling!\n");
		forward_gpu_gemm_ds(input, weights, output);
	}
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::forward_gpu_gemm_self(const Dtype* input,
	const Dtype* weights, Dtype* output)
{
	if (!conv_on_self_)
	{
		printf("Error: should be conv on self!\n");
		exit(0);
	}
	const Dtype* col_buff = input;
	if (!is_1x1_)
	{
		CBatchStructure self_struct_cpu;
		CGlobalSys<Dtype>::setBatchStructure(self_struct_cpu, m_top_struct_idx);	//top is the same as bottom

		int total_m = (int)self_struct_cpu.m_mSum_ptr[self_struct_cpu.m_batch_size];
		int total_def_num = (int)self_struct_cpu.m_defNumSum_ptr[self_struct_cpu.m_batch_size];

		CBatchStructure self_struct_GPU;
		CGlobalSys<Dtype>::setBatchStructure_GPU(self_struct_GPU, m_top_struct_idx);	//top is the same as bottom


		Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();

		////for debug
		//Blob<Dtype> tmp_buf;
		//tmp_buf.ReshapeLike(global_buf);
		//caffe::caffe_copy(tmp_buf.count(), global_buf.cpu_data(), tmp_buf.mutable_cpu_data());

#if RECORD_NEIGHBOR
		//hash2col_self_gpu<Dtype>(input, channels_, total_m, total_def_num,
		//	&self_struct_GPU, kernel_shape_.cpu_data(), global_buf.mutable_gpu_data(),nei_idx_.mutable_gpu_data());
		hash2col_self_gpu<Dtype>(input, conv_in_channels_, total_m, total_def_num,
			&self_struct_GPU, kernel_shape_.cpu_data(), global_buf.mutable_gpu_data(), nei_idx_.mutable_gpu_data());
#else
		//hash2col_self_gpu<Dtype>(input, channels_, total_m, total_def_num,
		//	&self_struct_GPU, kernel_shape_.cpu_data(), global_buf.mutable_gpu_data());
		hash2col_self_gpu<Dtype>(input, conv_in_channels_, total_m, total_def_num,
			&self_struct_GPU, kernel_shape_.cpu_data(), global_buf.mutable_gpu_data());
#endif
		col_buff = global_buf.gpu_data();

		////for debug
		//std::ofstream ofs("debug.txt");
		//for (int i = 0; i < global_buf.count(); i++)
		//{
		//	ofs << "CPU: " << tmp_buf.cpu_data()[i] << "; " << "GPU: " << global_buf.cpu_data()[i] << std::endl;
		//}
		//ofs.close();

	}
	caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, conv_out_channels_, conv_out_spatial_dim_, kernel_dim_,
		(Dtype)1., weights, col_buff,
		(Dtype)0., output);
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::forward_gpu_gemm_ds(const Dtype* input,
	const Dtype* weights, Dtype* output)
{
	if (conv_on_self_ || is_1x1_)
	{
		printf("Error: should be conv with stride>1!\n");
		exit(0);
	}
	
	
	CBatchStructure bt_struct_cpu, tp_struct_cpu;
	CGlobalSys<Dtype>::setBatchStructure(bt_struct_cpu, m_conv_bt_struct_idx);	
	int bt_total_m = (int)bt_struct_cpu.m_mSum_ptr[bt_struct_cpu.m_batch_size];
	int bt_total_def_num = (int)bt_struct_cpu.m_defNumSum_ptr[bt_struct_cpu.m_batch_size];
	CGlobalSys<Dtype>::setBatchStructure(tp_struct_cpu, m_conv_top_struct_idx);
	int tp_total_m = (int)tp_struct_cpu.m_mSum_ptr[tp_struct_cpu.m_batch_size];
	int tp_total_def_num = (int)tp_struct_cpu.m_defNumSum_ptr[tp_struct_cpu.m_batch_size];

	CBatchStructure bt_struct_GPU, tp_struct_GPU;
	CGlobalSys<Dtype>::setBatchStructure_GPU(bt_struct_GPU, m_conv_bt_struct_idx);
	CGlobalSys<Dtype>::setBatchStructure_GPU(tp_struct_GPU, m_conv_top_struct_idx);


	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();

	hash2col_ds_gpu<Dtype>(input, conv_in_channels_, tp_total_m, tp_total_def_num, bt_total_def_num,
		&bt_struct_GPU, &tp_struct_GPU, kernel_shape_.cpu_data(), stride_.cpu_data(), pad_.cpu_data(),global_buf.mutable_gpu_data());

	const Dtype* col_buff = global_buf.gpu_data();
	
	caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, conv_out_channels_, conv_out_spatial_dim_, kernel_dim_,
		(Dtype)1., weights, col_buff,
		(Dtype)0., output);
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::forward_gpu_bias(Dtype* output, const Dtype* bias)
{
	const Blob<Dtype> &bias_multiplier = CGlobalSys<Dtype>::get_multiplier();
	caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, num_output_,
		out_spatial_dim_, 1, (Dtype)1., bias, bias_multiplier.gpu_data(),
		(Dtype)1., output);
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::backward_gpu_gemm(const Dtype* output,
	const Dtype* weights, Dtype* input)
{
	if (conv_on_self_)
	{
		//printf("conv backward GPU on self!\n");
		backward_gpu_gemm_self(output, weights, input);
	}
	else
	{
		//printf("conv backward GPU with down sampling!\n");
		backward_gpu_gemm_ds(output, weights, input);
	}
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::backward_gpu_gemm_self(const Dtype* output,
	const Dtype* weights, Dtype* input)
{
	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
	Dtype* col_buff = global_buf.mutable_gpu_data();
	if (is_1x1_)
	{
		col_buff = input;
	}

	caffe_gpu_gemm<Dtype>(CblasTrans, CblasNoTrans, kernel_dim_,
		conv_out_spatial_dim_, conv_out_channels_,
		(Dtype)1., weights, output,
		(Dtype)0., col_buff);

	if (!is_1x1_)
	{
		CBatchStructure self_struct_cpu;
		CGlobalSys<Dtype>::setBatchStructure(self_struct_cpu, m_top_struct_idx);	//top is the same as bottom
		int total_m = (int)self_struct_cpu.m_mSum_ptr[self_struct_cpu.m_batch_size];
		int total_def_num = (int)self_struct_cpu.m_defNumSum_ptr[self_struct_cpu.m_batch_size];


		CBatchStructure self_struct_gpu;
		CGlobalSys<Dtype>::setBatchStructure_GPU(self_struct_gpu, m_top_struct_idx);	//top is the same as bottom

#if RECORD_NEIGHBOR
		//col2hash_self_gpu(input, channels_, total_m, total_def_num, &self_struct_gpu, kernel_shape_.cpu_data(),
		//	col_buff,nei_idx_.gpu_data());
		col2hash_self_gpu(input, conv_in_channels_, total_m, total_def_num, &self_struct_gpu, kernel_shape_.cpu_data(),
			col_buff, nei_idx_.gpu_data());
#else
		//col2hash_self_gpu(input, channels_, total_m, total_def_num,&self_struct_gpu, kernel_shape_.cpu_data(),
		//	col_buff);
		col2hash_self_gpu(input, conv_in_channels_, total_m, total_def_num, &self_struct_gpu, kernel_shape_.cpu_data(),
			col_buff);
#endif
	}
}


template <typename Dtype>
void BaseConvHashLayer<Dtype>::backward_gpu_gemm_ds(const Dtype* output,
	const Dtype* weights, Dtype* input)
{
	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
	Dtype* col_buff = global_buf.mutable_gpu_data();
	
	caffe_gpu_gemm<Dtype>(CblasTrans, CblasNoTrans, kernel_dim_,
		conv_out_spatial_dim_, conv_out_channels_,
		(Dtype)1., weights, output,
		(Dtype)0., col_buff);


	CBatchStructure bt_struct_cpu, top_struct_cpu;
	CGlobalSys<Dtype>::setBatchStructure(bt_struct_cpu, m_conv_bt_struct_idx);	
	CGlobalSys<Dtype>::setBatchStructure(top_struct_cpu, m_conv_top_struct_idx);
	int bt_total_m = (int)bt_struct_cpu.m_mSum_ptr[bt_struct_cpu.m_batch_size];
	int bt_def_num = (int)bt_struct_cpu.m_defNumSum_ptr[bt_struct_cpu.m_batch_size];
	int top_total_m = (int)top_struct_cpu.m_mSum_ptr[top_struct_cpu.m_batch_size];
	int top_def_num = (int)top_struct_cpu.m_defNumSum_ptr[top_struct_cpu.m_batch_size];


	CBatchStructure bt_struct_gpu, top_struct_gpu;
	CGlobalSys<Dtype>::setBatchStructure_GPU(bt_struct_gpu, m_conv_bt_struct_idx);	
	CGlobalSys<Dtype>::setBatchStructure_GPU(top_struct_gpu, m_conv_top_struct_idx);

	col2hash_ds_gpu(input, conv_in_channels_, top_total_m, top_def_num,bt_def_num, &bt_struct_gpu, &top_struct_gpu,kernel_shape_.cpu_data(), stride_.cpu_data(),pad_.cpu_data(),
		col_buff);
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::weight_gpu_gemm(const Dtype* input,
	const Dtype* output, Dtype* weights)
{
	if (conv_on_self_)
	{
		//printf("conv weight gemm GPU on self!\n");
		weight_gpu_gemm_self(input, output, weights);
	}
	else
	{
		//printf("conv weight gemm GPU with down sampling!\n");
		weight_gpu_gemm_ds(input, output, weights);
	}
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::weight_gpu_gemm_self(const Dtype* input,
	const Dtype* output, Dtype* weights)
{
	const Dtype* col_buff = input;

	CBatchStructure self_struct_cpu;
	CGlobalSys<Dtype>::setBatchStructure(self_struct_cpu, m_top_struct_idx);	//top is the same as bottom
	int total_m = (int)self_struct_cpu.m_mSum_ptr[self_struct_cpu.m_batch_size];
	int total_def_num = (int)self_struct_cpu.m_defNumSum_ptr[self_struct_cpu.m_batch_size];

	if (!is_1x1_)
	{
		CBatchStructure self_struct_GPU;
		CGlobalSys<Dtype>::setBatchStructure_GPU(self_struct_GPU, m_top_struct_idx);	//top is the same as bottom
		Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
		//hash2col_self_gpu(input, channels_, total_m, total_def_num, &self_struct_GPU, kernel_shape_.cpu_data(), global_buf.mutable_gpu_data());
		hash2col_self_gpu(input, conv_in_channels_, total_m, total_def_num, &self_struct_GPU, kernel_shape_.cpu_data(), global_buf.mutable_gpu_data());
		col_buff = global_buf.gpu_data();
	}


#ifdef USE_CUDNN
	// Cudnn
	// Interestingly, we find the GEMM is nearly ten times slower than Cudnn
	// when doing weight back-propagation. It's possible because the ill matrix
	// size (suach as A(1000, 120000) * B(120000, 500) = W(1000, 500)).
	const int *kernel_shape_data = kernel_shape_.cpu_data();
	int kernel_spatial_dim = kernel_shape_data[0] * kernel_shape_data[1] * kernel_shape_data[2];
	cudnn::setTensor4dDesc<Dtype>(&bottom_desc_, 1, conv_in_channels_,
		kernel_spatial_dim, total_def_num);
	cudnn::setTensor4dDesc<Dtype>(&top_desc_, 1, conv_out_channels_,
		1, total_def_num);
	cudnn::setConvolutionDesc<Dtype>(&conv_desc_, bottom_desc_, filter_desc_,
		0, 0, 1, 1);
	// choose backward algorithm for filter
	bwd_filter_algo_ = CUDNN_CONVOLUTION_BWD_FILTER_ALGO_3;
	//CUDNN_CHECK(cudnnGetConvolutionBackwardFilterAlgorithm(handle_,
	//	bottom_desc_, top_desc_, conv_desc_, filter_desc_,
	//	CUDNN_CONVOLUTION_BWD_FILTER_SPECIFY_WORKSPACE_LIMIT,
	//	filter_workspace_size_, &bwd_filter_algo_));
	CUDNN_CHECK(cudnnConvolutionBackwardFilter(handle_,
		cudnn::dataType<Dtype>::one, bottom_desc_, col_buff, top_desc_, output,
		conv_desc_, bwd_filter_algo_, filter_workspace_.mutable_gpu_data(),
		filter_workspace_size_, cudnn::dataType<Dtype>::zero, filter_desc_,
		weights));
#else
	// GEMM
	caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasTrans, conv_out_channels_,
		kernel_dim_, conv_out_spatial_dim_,
		(Dtype)1., output, col_buff,
		(Dtype)0., weights);
#endif // USE_CUDNN
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::weight_gpu_gemm_ds(const Dtype* input,
	const Dtype* output, Dtype* weights)
{
	const Dtype* col_buff = input;

	CBatchStructure bt_struct_cpu, top_struct_cpu;
	CGlobalSys<Dtype>::setBatchStructure(bt_struct_cpu, m_conv_bt_struct_idx);
	CGlobalSys<Dtype>::setBatchStructure(top_struct_cpu, m_conv_top_struct_idx);
	int bt_total_m = (int)bt_struct_cpu.m_mSum_ptr[bt_struct_cpu.m_batch_size];
	int bt_total_def_num = (int)bt_struct_cpu.m_defNumSum_ptr[bt_struct_cpu.m_batch_size];
	int top_total_m = (int)top_struct_cpu.m_mSum_ptr[top_struct_cpu.m_batch_size];
	int top_total_def_num = (int)top_struct_cpu.m_defNumSum_ptr[top_struct_cpu.m_batch_size];

	CBatchStructure bt_struct_GPU, top_struct_GPU;
	CGlobalSys<Dtype>::setBatchStructure_GPU(bt_struct_GPU, m_conv_bt_struct_idx);	
	CGlobalSys<Dtype>::setBatchStructure_GPU(top_struct_GPU, m_conv_top_struct_idx);
	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
	hash2col_ds_gpu(input, conv_in_channels_, top_total_m, top_total_def_num, bt_total_def_num, &bt_struct_GPU, &top_struct_GPU, kernel_shape_.cpu_data(), 
		stride_.cpu_data(), pad_.cpu_data(),global_buf.mutable_gpu_data());
	col_buff = global_buf.gpu_data();
	

#ifdef USE_CUDNN
	// Cudnn
	// Interestingly, we find the GEMM is nearly ten times slower than Cudnn
	// when doing weight back-propagation. It's possible because the ill matrix
	// size (suach as A(1000, 120000) * B(120000, 500) = W(1000, 500)).
	const int *kernel_shape_data = kernel_shape_.cpu_data();
	int kernel_spatial_dim = kernel_shape_data[0] * kernel_shape_data[1] * kernel_shape_data[2];
	cudnn::setTensor4dDesc<Dtype>(&bottom_desc_, 1, conv_in_channels_,
		kernel_spatial_dim, top_total_def_num);
	cudnn::setTensor4dDesc<Dtype>(&top_desc_, 1, conv_out_channels_,
		1, top_total_def_num);
	cudnn::setConvolutionDesc<Dtype>(&conv_desc_, bottom_desc_, filter_desc_,
		0, 0, 1, 1);
	// choose backward algorithm for filter
	bwd_filter_algo_ = CUDNN_CONVOLUTION_BWD_FILTER_ALGO_3;
	//CUDNN_CHECK(cudnnGetConvolutionBackwardFilterAlgorithm(handle_,
	//	bottom_desc_, top_desc_, conv_desc_, filter_desc_,
	//	CUDNN_CONVOLUTION_BWD_FILTER_SPECIFY_WORKSPACE_LIMIT,
	//	filter_workspace_size_, &bwd_filter_algo_));
	CUDNN_CHECK(cudnnConvolutionBackwardFilter(handle_,
		cudnn::dataType<Dtype>::one, bottom_desc_, col_buff, top_desc_, output,
		conv_desc_, bwd_filter_algo_, filter_workspace_.mutable_gpu_data(),
		filter_workspace_size_, cudnn::dataType<Dtype>::zero, filter_desc_,
		weights));
#else
	// GEMM
	caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasTrans, conv_out_channels_,
		kernel_dim_, conv_out_spatial_dim_,
		(Dtype)1., output, col_buff,
		(Dtype)0., weights);
#endif // USE_CUDNN
}

template <typename Dtype>
void BaseConvHashLayer<Dtype>::backward_gpu_bias(Dtype* bias,
	const Dtype* input)
{
	Blob<Dtype> &bias_multiplier = CGlobalSys<Dtype>::get_multiplier();

	caffe_gpu_gemv<Dtype>(CblasNoTrans, num_output_, out_spatial_dim_, 1.,
		input, bias_multiplier.gpu_data(), (Dtype)0., bias);
}
#endif
#if 0

template <typename Dtype>
int BaseConvHashLayer<Dtype>::writeDenseKernel_2_HF5(const char *filename) 
{
	if (blobs_.size()<1)
	{
		return 0;
	}
	writeDense_2_HF5((const float*)blobs_[0]->cpu_data(), 
		blobs_[0]->shape(0), blobs_[0]->shape(2), blobs_[0]->shape(1), filename);
	return 1;
}

template <typename Dtype>
int BaseConvHashLayer<Dtype>::writeBias_2_HF5(const char *filename)
{
	if (blobs_.size() < 2)
	{
		return 0;
	}
	writeDense_2_HF5((const float*)blobs_[1]->cpu_data(),
		1, 1, blobs_[1]->shape(0), filename);
	return 1;
}



template <typename Dtype>
int BaseConvHashLayer<Dtype>::writeDenseKernelDif_2_HF5(const char *filename)
{
	if (blobs_.size() < 1)
	{
		return 0;
	}
	writeDense_2_HF5((const float*)blobs_[0]->cpu_diff(),
		blobs_[0]->shape(0), blobs_[0]->shape(2), blobs_[0]->shape(1), filename);
	return 1;
}

template <typename Dtype>
int BaseConvHashLayer<Dtype>::writeBiasDif_2_HF5(const char *filename)
{
	if (blobs_.size() < 2)
	{
		return 0;
	}
	writeDense_2_HF5((const float*)blobs_[1]->cpu_diff(),
		1, 1, blobs_[1]->shape(0), filename);
	return 1;
}


#endif  // !CPU_ONLY



INSTANTIATE_CLASS(BaseConvHashLayer);



}  // namespace caffe
