#include <algorithm>
#include <vector>

#include "caffe/layers/bn_hash_layer.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/util/HashData.h"

namespace caffe {

template <typename Dtype>
void BNHashLayer<Dtype>::LayerSetUp(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) 
{
	//We do not allow bottom and top share the same memory in this implementation
	if (bottom[HASH_DATA_BLOB] == top[HASH_DATA_BLOB])
	{
		LOG(INFO) << "Fatal error: We do not allow bottom and top share the same memory in this implementation!";
		exit(0);
	}

  const BNHashParameter &param = this->layer_param_.bn_hash_param();
  moving_average_fraction_ = param.moving_average_fraction();
  use_global_stats_ = this->phase_ == TEST;
  if (param.has_use_global_stats())
    use_global_stats_ = param.use_global_stats();
  disable_mean_ = param.disable_mean();
  disable_variance_ = param.disable_variance();
  channels_ = bottom[HASH_DATA_BLOB]->shape(CHANNEL_AXIS);
  eps_ = param.eps();
  if (this->blobs_.size() > 0) {
	  LOG(INFO) << "Skipping parameter initialization";
  }
  else {
	  this->blobs_.resize(3);
	  vector<int> sz;
	  sz.push_back(channels_);
	  this->blobs_[0].reset(new Blob<Dtype>(sz));
	  this->blobs_[1].reset(new Blob<Dtype>(sz));
	  sz[0] = 1;
	  this->blobs_[2].reset(new Blob<Dtype>(sz));
	  for (int i = 0; i < 3; ++i) {
		  caffe_set(this->blobs_[i]->count(), Dtype(0),
			  this->blobs_[i]->mutable_cpu_data());
	  }
  }
  //if (param.engine() != BatchNormParameter_Engine_CUDNN) 
  {
    // Mask statistics from optimization by setting local learning rates
    // for mean, variance, and the bias correction to zero.
    for (int i = 0; i < this->blobs_.size(); ++i) {
      if (this->layer_param_.param_size() == i) {
        ParamSpec* fixed_param_spec = this->layer_param_.add_param();
        fixed_param_spec->set_lr_mult(0.f);
      }
      else {
        CHECK_EQ(this->layer_param_.param(i).lr_mult(), 0.f)
          << "Cannot configure batch normalization statistics as layer "
          << "parameters.";
      }
    }
  }
}



template <typename Dtype>
void BNHashLayer<Dtype>::reshape_topHashData(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{	
	top[HASH_DATA_BLOB]->ReshapeLike(*bottom[HASH_DATA_BLOB]);
}


template <typename Dtype>
void BNHashLayer<Dtype>::Reshape(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{
	reshape_topHashData(bottom, top);

	vector<int> sz;
	sz.push_back(channels_);
	mean_.Reshape(sz);
	variance_.Reshape(sz);
	inv_sqrt_var_.Reshape(sz);
	
	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
	global_buf.ReshapeLike(*bottom[HASH_DATA_BLOB]);

	// a multipler for Mv
	Blob<Dtype> &multiplier = CGlobalSys<Dtype>::get_multiplier();
	sz[0] = bottom[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
	multiplier.Reshape(sz);
	switch (Caffe::mode())
	{
	default:
		break;
	case Caffe::GPU:
		caffe_gpu_set(multiplier.count(), Dtype(1), multiplier.mutable_gpu_data());
		break;
	case Caffe::CPU:
		caffe_set(multiplier.count(), Dtype(1), multiplier.mutable_cpu_data());
		break;
	}
}


template <typename Dtype>
void BNHashLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{
	//total num
	const int total_defNum = bottom[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
	const Dtype mean_div = Dtype(1) / Dtype(total_defNum);
	//const Dtype var_div = Dtype(1) / Dtype(std::max(1, total_defNum - 1));
	const Dtype var_div = mean_div;	//will be bias-corrected when adding to blob[1]

	
	caffe_copy(bottom[0]->count(), bottom[HASH_DATA_BLOB]->cpu_data(), top[HASH_DATA_BLOB]->mutable_cpu_data());

	const Blob<Dtype> &multiplier = CGlobalSys<Dtype>::get_multiplier();
	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();

	if (use_global_stats_) {
		// use the stored mean/variance estimates.
		const Dtype scale_factor = this->blobs_[2]->cpu_data()[0] == 0 ?
			0 : 1 / this->blobs_[2]->cpu_data()[0];
		caffe_cpu_scale(variance_.count(), scale_factor,
			this->blobs_[0]->cpu_data(), mean_.mutable_cpu_data());
		caffe_cpu_scale(variance_.count(), scale_factor,
			this->blobs_[1]->cpu_data(), variance_.mutable_cpu_data());
	}
	else
	{
		if (!disable_mean_) {
			/************ 1. compute the mean EX for each channel *************/
			caffe_cpu_gemv(CblasNoTrans, channels_, total_defNum, mean_div,
				bottom[HASH_DATA_BLOB]->cpu_data(), multiplier.cpu_data(), Dtype(0), mean_.mutable_cpu_data());
		}
	}


	/**********************2 substract mean****************/
	if (!disable_mean_) {
		// subtract mean
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_,
			total_defNum, 1, (Dtype)-1, mean_.cpu_data(), multiplier.cpu_data(),
			(Dtype)1, top[HASH_DATA_BLOB]->mutable_cpu_data());
	}
	/********************3. compute variance using var(X) = E((X-EX)^2)***********************/
	if (!use_global_stats_)
	{
		if (!disable_variance_) {
			caffe_powx(top[HASH_DATA_BLOB]->count(), top[HASH_DATA_BLOB]->cpu_data(), Dtype(2),
				global_buf.mutable_cpu_data());  // (X-EX)^2
			caffe_cpu_gemv(CblasNoTrans, channels_, total_defNum, var_div,
				global_buf.cpu_data(), multiplier.cpu_data(), Dtype(0), variance_.mutable_cpu_data());// E((X_EX)^2)
		}

		// compute and save moving average
		this->blobs_[2]->mutable_cpu_data()[0] *= moving_average_fraction_;
		this->blobs_[2]->mutable_cpu_data()[0] += 1;
		if (!disable_mean_) {
			caffe_cpu_axpby(mean_.count(), Dtype(1), mean_.cpu_data(),
				moving_average_fraction_, this->blobs_[0]->mutable_cpu_data());
		}
		if (!disable_variance_) {
			Dtype bias_correction_factor = total_defNum > 1 ? Dtype(total_defNum) / Dtype(total_defNum - 1) : 1;
			caffe_cpu_axpby(variance_.count(), bias_correction_factor,
				variance_.cpu_data(), moving_average_fraction_, this->blobs_[1]->mutable_cpu_data());
		}
	}

	/********************4. compute final top (X-mean(X))/(sqrt(var(X)+eps))***********************/
	if (!disable_variance_) {
		// normalize variance
		caffe_add_scalar(variance_.count(), eps_, variance_.mutable_cpu_data());
		caffe_powx(variance_.count(), variance_.cpu_data(), Dtype(0.5),
			variance_.mutable_cpu_data());

#if 0
		// replicate variance to input size
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_, total_defNum, 1,
			(Dtype)1, variance_.cpu_data(), multiplier.cpu_data(), (Dtype)0,
			global_buf.mutable_cpu_data());

		caffe_div(global_buf.count(), top[HASH_DATA_BLOB]->cpu_data(), global_buf.cpu_data(), top[HASH_DATA_BLOB]->mutable_cpu_data());
#else
		//set variance to inV
		for (int c = 0; c < channels_; c++)
		{
			inv_sqrt_var_.mutable_cpu_data()[c] = 1.f / variance_.cpu_data()[c];
		}
		// replicate inv_variance to input size
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_, total_defNum, 1,
			(Dtype)1, inv_sqrt_var_.cpu_data(), multiplier.cpu_data(), (Dtype)0,
			global_buf.mutable_cpu_data());

		caffe_mul(global_buf.count(), top[HASH_DATA_BLOB]->cpu_data(), global_buf.cpu_data(), top[HASH_DATA_BLOB]->mutable_cpu_data());
#endif
	}
}

template <typename Dtype>
void BNHashLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
    const vector<bool>& propagate_down,
    const vector<Blob<Dtype>*>& bottom) 
{
	const Dtype* top_diff = top[HASH_DATA_BLOB]->cpu_diff();
	const Dtype* top_data = top[HASH_DATA_BLOB]->cpu_data();
	Dtype* bottom_diff = bottom[HASH_DATA_BLOB]->mutable_cpu_diff();
	const int total_defNum = top[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
	const Dtype mean_div = Dtype(1) / Dtype(total_defNum);

	const Blob<Dtype> &multiplier = CGlobalSys<Dtype>::get_multiplier();
	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();

	if (use_global_stats_) 
	{
		if (disable_variance_) {
			caffe_copy(top[HASH_DATA_BLOB]->count(), top_diff, bottom_diff);
		}
		else
		{
			// replicate inv_variance to input size
			caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_, total_defNum, 1,
				(Dtype)1, inv_sqrt_var_.cpu_data(), multiplier.cpu_data(), (Dtype)0,
				global_buf.mutable_cpu_data());
			caffe_mul(top[HASH_DATA_BLOB]->count(), top_diff, global_buf.cpu_data(), bottom_diff);
		}
		return;
	}


	// if Y = (X-mean(X))/(sqrt(var(X)+eps)), then
	//
	// dE(Y)/dX =
	//   (dE/dY - mean(dE/dY) - mean(dE/dY \cdot Y) \cdot Y)
	//     ./ sqrt(var(X) + eps)
	//
	// where \cdot and ./ are hadamard product and elementwise division,
	// respectively, dE/dY is the top diff, and mean/var/sum are all computed
	// along all dimensions except the channels dimension.  In the above
	// equation, the operations allow for expansion (i.e. broadcast) along all
	// dimensions except the channels dimension where required.
	// --------------------------------------------------------
	// If disable_vairance is set, the derivative change to
	// dE(Y)/dX = dE/dY - mean(dE/dY)
	// If disable_mean is set, derivative becomes
	// dE(Y)/dX =
	//   (dE/dY - mean(dE/dY \cdot Y) \cdot Y)
	//     ./ sqrt(var(X) + eps)

	if (!disable_variance_) {
		//step1. mean(dE/dY \cdot Y)
		//dE/dY \cdot Y;
		caffe_mul(top[HASH_DATA_BLOB]->count(), top_data, top_diff, global_buf.mutable_cpu_data());
		//mean
		caffe_cpu_gemv(CblasNoTrans, channels_, total_defNum, mean_div,
			global_buf.cpu_data(), multiplier.cpu_data(), Dtype(0), mean_.mutable_cpu_data());

		//step2. mean(dE/dY \cdot Y) \cdot Y
		//reshape mean to input size
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_, total_defNum, 1,
			(Dtype)1, mean_.cpu_data(), multiplier.cpu_data(), (Dtype)0,
			global_buf.mutable_cpu_data());
		// mean(dE/dY \cdot Y) \cdot Y
		caffe_mul(top[HASH_DATA_BLOB]->count(), top_data, global_buf.cpu_data(), global_buf.mutable_cpu_data());

		//dE/dY - mean(dE/dY \cdot Y) \cdot Y
		caffe_sub(top[HASH_DATA_BLOB]->count(), top_diff, global_buf.cpu_data(), bottom_diff);
	}
	else {
		caffe_copy(top[HASH_DATA_BLOB]->count(), top_diff, bottom_diff);
	}


	//step3. dE/dY - mean(dE/dY) - mean(dE/dY \cdot Y) \cdot Y
	if (!disable_mean_) {
		//mean(dE/dY)
		caffe_cpu_gemv(CblasNoTrans, channels_, total_defNum, mean_div,
			top_diff, multiplier.cpu_data(), Dtype(0), mean_.mutable_cpu_data());

		//-= mean(dE/dY)
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_,
			total_defNum, 1, (Dtype)-1, mean_.cpu_data(), multiplier.cpu_data(),
			(Dtype)1, bottom_diff);
	}


	if (!disable_variance_) {
		//step4. (dE/dY - mean(dE/dY) - mean(dE/dY \cdot Y) \cdot Y) / sqrt(var(X) + eps)
		// replicate inv_variance to input size
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_, total_defNum, 1,
			(Dtype)1, inv_sqrt_var_.cpu_data(), multiplier.cpu_data(), (Dtype)0,
			global_buf.mutable_cpu_data());
		caffe_mul(top[HASH_DATA_BLOB]->count(), bottom_diff, global_buf.cpu_data(), bottom_diff);
	}
}


#ifdef CPU_ONLY
STUB_GPU(BNHashLayer);
#endif

INSTANTIATE_CLASS(BNHashLayer);
REGISTER_LAYER_CLASS(BNHash);
}  // namespace caffe
