#include <algorithm>
#include <vector>

#include "caffe/layers/bn_hash_layer.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe\util\HashData.h"

namespace caffe {
	template <typename Dtype>
	__global__ void substract_mean_kernel(Dtype* temp, int channels, int total_defNum, const Dtype* mean)
	{
		CUDA_KERNEL_LOOP(index, channels*total_defNum)
		{
			const int c = index / total_defNum;
			temp[index] -= mean[c];
		}
	}

#if 0
	template <typename Dtype>
	__global__ void inv_sqrt_eps_var_kernel(Dtype* temp, int channels, int total_defNum, const Dtype* var, Dtype eps)
	{
		CUDA_KERNEL_LOOP(index, channels*total_defNum)
		{
			const int c = index / total_defNum;
			temp[index] /= sqrt(var[c] + eps);
		}
	}

	template <typename Dtype>
	__global__ void invSqrtVar_kernel(const Dtype* var, Dtype *inv_sqrt_var, int channels, Dtype eps)
	{
		CUDA_KERNEL_LOOP(index, channels)
		{
			inv_sqrt_var[index] = 1.f / sqrt(var[index] + eps);
		}
	}
#else
	template <typename Dtype>
	__global__ void inv_sqrt_eps_var_kernel(Dtype* temp, int channels, int total_defNum, const Dtype* var, Dtype *inv_sqrt_var, Dtype eps)
	{
		CUDA_KERNEL_LOOP(index, channels*total_defNum)
		{
			const int c = index / total_defNum;
			inv_sqrt_var[c] = 1.f / sqrt(var[c] + eps);
			temp[index] *= inv_sqrt_var[c];
		}
	}
#endif


	template <typename Dtype>
	void BNHashLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
		const vector<Blob<Dtype>*>& top)
	{
		const Dtype* bottom_data = bottom[HASH_DATA_BLOB]->gpu_data();
		Dtype* top_data = top[HASH_DATA_BLOB]->mutable_gpu_data();
		//total num
		const int total_defNum = bottom[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
		const Dtype mean_div = Dtype(1) / Dtype(total_defNum);
		//const Dtype var_div = Dtype(1) / Dtype(std::max(1, total_defNum - 1));
		const Dtype var_div = mean_div;	//will be bias-corrected when adding to blob[1]


		caffe_copy(bottom[HASH_DATA_BLOB]->count(), bottom_data, top_data);

		const Blob<Dtype> &multiplier = CGlobalSys<Dtype>::get_multiplier();
		Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();

		if (use_global_stats_) 
		{
			// use the stored mean/variance estimates.
			const Dtype scale_factor = this->blobs_[2]->cpu_data()[0] == 0 ?
				0 : 1 / this->blobs_[2]->cpu_data()[0];
			caffe_gpu_scale(variance_.count(), scale_factor,
				this->blobs_[0]->gpu_data(), mean_.mutable_gpu_data());
			caffe_gpu_scale(variance_.count(), scale_factor,
				this->blobs_[1]->gpu_data(), variance_.mutable_gpu_data());
		}
		else
		{
			/********1. compute the mean EX for each channel *************/
			if (!disable_mean_) {
				// compute mean
				caffe_gpu_gemv(CblasNoTrans, channels_, total_defNum, mean_div,
					bottom_data, multiplier.gpu_data(), Dtype(0), mean_.mutable_gpu_data());
			}
		}

		if (!disable_mean_) {
			/**********************2 substract mean****************/
			substract_mean_kernel << <CAFFE_GET_BLOCKS(channels_*total_defNum), CAFFE_CUDA_NUM_THREADS >> > (
				top_data, channels_, total_defNum, mean_.gpu_data()
				);
		}

		/********************3. compute variance using var(X) = E((X-EX)^2)***********************/
		if (!use_global_stats_)
		{
			if (!disable_variance_) {
				caffe_gpu_mul(bottom[HASH_DATA_BLOB]->count(), top_data, top_data, global_buf.mutable_gpu_data());// (X-EX)^2
				caffe_gpu_gemv(CblasNoTrans, channels_, total_defNum, var_div,
					global_buf.gpu_data(), multiplier.gpu_data(), Dtype(0), variance_.mutable_gpu_data());// E((X_EX)^2)
			}

			// compute and save moving average
			this->blobs_[2]->mutable_cpu_data()[0] *= moving_average_fraction_;
			this->blobs_[2]->mutable_cpu_data()[0] += 1;

			if (!disable_mean_) {
				caffe_gpu_axpby(mean_.count(), Dtype(1), mean_.gpu_data(),
					moving_average_fraction_, this->blobs_[0]->mutable_gpu_data());
			}
			if (!disable_variance_) {
				Dtype bias_correction_factor = total_defNum > 1 ? Dtype(total_defNum) / (total_defNum - 1) : 1;
				caffe_gpu_axpby(variance_.count(), bias_correction_factor,
					variance_.gpu_data(), moving_average_fraction_, this->blobs_[1]->mutable_gpu_data());
			}
		}

		/********************4. compute final top (X-mean(X))/(sqrt(var(X)+eps))***********************/
		if (!disable_variance_) {
#if 1
			// div by sqrt(var(X)+eps), and record inv_sqrt_var for bp
			inv_sqrt_eps_var_kernel << <CAFFE_GET_BLOCKS(channels_*total_defNum), CAFFE_CUDA_NUM_THREADS >> > (
				top_data, channels_, total_defNum, variance_.gpu_data(), inv_sqrt_var_.mutable_gpu_data(),eps_
				);
#else
			// normalize variance
			// div by sqrt(var(X)+eps)
			inv_sqrt_eps_var_kernel << <CAFFE_GET_BLOCKS(channels_*total_defNum), CAFFE_CUDA_NUM_THREADS >> > (
				top_data, channels_, total_defNum, variance_.gpu_data(), eps_
				);
			//set variance to inV, which is used in BP
			//for (int c = 0; c < channels_; c++)
			//{
			//	//inv_sqrt_var_.mutable_cpu_data()[c] = 1.f / variance_.cpu_data()[c];
			//	inv_sqrt_var_.mutable_cpu_data()[c] = 1.f / sqrt(variance_.cpu_data()[c] + eps_);
			//	//printf("inv sqrt var %.6f\n", inv_sqrt_var_.cpu_data()[c]);
			//}
			invSqrtVar_kernel << <CAFFE_GET_BLOCKS(channels_), CAFFE_CUDA_NUM_THREADS >> > (
				variance_.gpu_data(), inv_sqrt_var_.mutable_gpu_data(), channels_, eps_
				);
#endif
		}
	}

	template <typename Dtype>
	void BNHashLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
		const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom)
	{
		const Dtype* top_diff = top[HASH_DATA_BLOB]->gpu_diff();
		const Dtype* top_data = top[HASH_DATA_BLOB]->gpu_data();
		Dtype* bottom_diff = bottom[HASH_DATA_BLOB]->mutable_gpu_diff();
		
		const int total_defNum = top[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
		const Dtype mean_div = Dtype(1) / Dtype(total_defNum);

		const Blob<Dtype> &multiplier = CGlobalSys<Dtype>::get_multiplier();
		Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();

		if (use_global_stats_)
		{
			if (disable_variance_) {
				caffe_copy(top[HASH_DATA_BLOB]->count(), top_diff, bottom_diff);
			}
			else
			{
				// replicate inv_variance to input size
				caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_, total_defNum, 1,
					(Dtype)1, inv_sqrt_var_.gpu_data(), multiplier.gpu_data(), (Dtype)0,
					global_buf.mutable_gpu_data());
				caffe_gpu_mul(top[HASH_DATA_BLOB]->count(), top_diff, global_buf.gpu_data(), bottom_diff);
			}
			return;
		}


		// if Y = (X-mean(X))/(sqrt(var(X)+eps)), then
		//
		// dE(Y)/dX =
		//   (dE/dY - mean(dE/dY) - mean(dE/dY \cdot Y) \cdot Y)
		//     ./ sqrt(var(X) + eps)
		//
		// where \cdot and ./ are hadamard product and elementwise division,
		// respectively, dE/dY is the top diff, and mean/var/sum are all computed
		// along all dimensions except the channels dimension.  In the above
		// equation, the operations allow for expansion (i.e. broadcast) along all
		// dimensions except the channels dimension where required.
		// --------------------------------------------------------
		// If disable_vairance is set, the derivative change to
		// dE(Y)/dX = dE/dY - mean(dE/dY)
		// If disable_mean is set, derivative becomes
		// dE(Y)/dX =
		//   (dE/dY - mean(dE/dY \cdot Y) \cdot Y)
		//     ./ sqrt(var(X) + eps)

#if 1
		if (!disable_variance_) {
			//step1. mean(dE/dY \cdot Y)
			//dE/dY \cdot Y; // NOTE: here temp_ is modified
			caffe_gpu_mul(top[HASH_DATA_BLOB]->count(), top_diff, top_data, global_buf.mutable_gpu_data());
			//mean
			caffe_gpu_gemv(CblasNoTrans, channels_, total_defNum, mean_div,
				global_buf.gpu_data(), multiplier.gpu_data(), Dtype(0), mean_.mutable_gpu_data());

			//step2. mean(dE/dY \cdot Y) \cdot Y
			//reshape mean to input size
			caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_, total_defNum, 1,
				(Dtype)1, mean_.gpu_data(), multiplier.gpu_data(), (Dtype)0,
				global_buf.mutable_gpu_data());
			// mean(dE/dY \cdot Y) \cdot Y
			caffe_gpu_mul(top[HASH_DATA_BLOB]->count(), global_buf.gpu_data(), top_data, global_buf.mutable_gpu_data());

			//dE/dY - mean(dE/dY \cdot Y) \cdot Y
			caffe_gpu_sub(top[HASH_DATA_BLOB]->count(), top_diff, global_buf.gpu_data(), bottom_diff);
		}
		else {
			caffe_copy(top[HASH_DATA_BLOB]->count(), top_diff, bottom_diff);
		}


		//step3. dE/dY - mean(dE/dY) - mean(dE/dY \cdot Y) \cdot Y
		if (!disable_mean_) {
			//mean(dE/dY)
			caffe_gpu_gemv(CblasNoTrans, channels_, total_defNum, mean_div,
				top_diff, multiplier.gpu_data(), Dtype(0), mean_.mutable_gpu_data());

			//-= mean(dE/dY)
			caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_,
				total_defNum, 1, (Dtype)-1, mean_.gpu_data(), multiplier.gpu_data(),
				(Dtype)1, bottom_diff);
		}

		if (!disable_variance_) {
			//step4. (dE/dY - mean(dE/dY) - mean(dE/dY \cdot Y) \cdot Y) / sqrt(var(X) + eps)
			// replicate inv_variance to input size
			caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_, total_defNum, 1,
				(Dtype)1, inv_sqrt_var_.gpu_data(), multiplier.gpu_data(), (Dtype)0,
				global_buf.mutable_gpu_data());
			caffe_gpu_mul(top[HASH_DATA_BLOB]->count(), bottom_diff, global_buf.gpu_data(), bottom_diff);
		}
#else
		//step1. mean(dE/dY \cdot Y)
		//dE/dY \cdot Y; // NOTE: here temp_ is modified
		caffe_gpu_mul(top[HASH_DATA_BLOB]->count(), top_diff, top_data, global_buf.mutable_gpu_data());
		//mean
		caffe_gpu_gemv(CblasNoTrans, channels_, total_defNum, mean_div,
			global_buf.gpu_data(), multiplier.gpu_data(), Dtype(0), mean_.mutable_gpu_data());

		//step2. mean(dE/dY \cdot Y) \cdot Y
		//reshape mean to input size
		caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_, total_defNum, 1,
			(Dtype)1, mean_.gpu_data(), multiplier.gpu_data(), (Dtype)0,
			global_buf.mutable_gpu_data());
		// mean(dE/dY \cdot Y) \cdot Y
		caffe_gpu_mul(top[HASH_DATA_BLOB]->count(), global_buf.gpu_data(), top_data, global_buf.mutable_gpu_data());

		//step3. dE/dY - mean(dE/dY) - mean(dE/dY \cdot Y) \cdot Y
		//mean(dE/dY)
		caffe_gpu_gemv(CblasNoTrans, channels_, total_defNum, mean_div,
			top_diff, multiplier.gpu_data(), Dtype(0), mean_.mutable_gpu_data());

		//dE/dY - mean(dE/dY \cdot Y) \cdot Y
		caffe_gpu_sub(top[HASH_DATA_BLOB]->count(), top_diff, global_buf.gpu_data(), bottom_diff);

		//-= mean(dE/dY)
		caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_,
			total_defNum, 1, (Dtype)-1, mean_.gpu_data(), multiplier.gpu_data(),
			(Dtype)1, bottom_diff);

		//step4. (dE/dY - mean(dE/dY) - mean(dE/dY \cdot Y) \cdot Y) / sqrt(var(X) + eps)
		// replicate inv_variance to input size
		caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, channels_, total_defNum, 1,
			(Dtype)1, inv_sqrt_var_.gpu_data(), multiplier.gpu_data(), (Dtype)0,
			global_buf.mutable_gpu_data());
		caffe_gpu_mul(top[HASH_DATA_BLOB]->count(), bottom_diff, global_buf.gpu_data(), bottom_diff);
#endif
	}

	INSTANTIATE_LAYER_GPU_FUNCS(BNHashLayer);
}  // namespace caffe
