#include <fstream>  // NOLINT(readability/streams)
#include <string>
#include <vector>

#include "stdint.h"

#include "caffe/layers/hash_data_layer.hpp"
#include "caffe/util/MyMacro.h"
#include <boost/thread.hpp>

#define MULT_THREAD 1

namespace caffe {

	template <typename Dtype>
	HashDataLayer<Dtype>::HashDataLayer(const LayerParameter& param)
		: Layer<Dtype>(param), offset_()
		,prefetch_(param.data_param().prefetch()),prefetch_free_(), prefetch_full_(), prefetch_current_(NULL) 
	{
#if MULT_THREAD
		for (int i = 0; i < prefetch_.size(); ++i) 
		{
			prefetch_[i].reset(new HierHash_Label_HashLabel_Batch<Dtype>());
			prefetch_free_.push(prefetch_[i].get());
		}
#endif
		segment_ = false;
	}

template <typename Dtype>
HashDataLayer<Dtype>::~HashDataLayer<Dtype>() 
{
#if MULT_THREAD
	this->StopInternalThread();
#endif
	destroyHierHashes();
	destroyBatch();
	destroyHashLabels();
	destroyBatchHashLabels();
}

template <typename Dtype>
void HashDataLayer<Dtype>::destroyHierHashes() 
{
	for (int i=0;i<(int)m_vpHierHashes.size();i++)
	{
		delete m_vpHierHashes[i];
	}
	m_vpHierHashes.resize(0);
}

template <typename Dtype>
void HashDataLayer<Dtype>::destroyBatch()
{
	for (int i = 0; i<(int)m_hierHash_batch.size(); i++)
	{
		delete m_hierHash_batch[i];
	}
	m_hierHash_batch.resize(0);
}

template <typename Dtype>
int HashDataLayer<Dtype>::loadHierHashes(std::vector<CHierarchyHash *> &vpHierHashes, FILE *fp)
{
	int n = 0;
	fread(&n, sizeof(int), 1, fp);

	if (n <= 0)
	{
		printf("Fatal error: hier hash num <= 0!\n");
		exit(0);
	}

	//actually when load batches, n will always to the same
	int ori_n = (int)vpHierHashes.size();
	if (ori_n<n)
	{
		for (int i = ori_n; i<n; i++)
		{
			vpHierHashes.push_back(new CHierarchyHash());
		}
	}
	if (ori_n>n)
	{
		for (int i = ori_n - 1; i >= n; i--)
		{
			delete vpHierHashes[i];
			vpHierHashes.pop_back();
		}
	}

	for (int i = 0; i < n; i++)
	{
		vpHierHashes[i]->load(fp);
	}
	return n;
}


template <typename Dtype>
bool HashDataLayer<Dtype>::Load_labels(Blob<Dtype> &label_blob, FILE *fp)
{
	int hash_n;
	fread(&hash_n, sizeof(int), 1, fp);
	int label_size;
	fread(&label_size, sizeof(int), 1, fp);

	if (label_size < 1 || hash_n < 1)
	{
		LOG(INFO)<<"Error: invalid hash num "<<hash_n<<" or label size "<<label_size;
		return false;
	}

	std::vector<int> label_shape;
	label_shape.push_back(hash_n);
	label_shape.push_back(label_size);
	label_blob.Reshape(label_shape);
	fread(label_blob.mutable_cpu_data(), sizeof(Dtype), hash_n*label_size, fp);

	return true;
}


// Load data and label from Hash filename into the class property blobs.
template <typename Dtype>
void HashDataLayer<Dtype>::LoadHashLabelFileData(const char* filename) 
{
  //NOTE: we dont destroy the memory to avoid the memory cost; We only enlarge the memory if needed
  //destroyHashes(m_hashes);
  LOG(INFO) << "Loading HashLabel file: " << filename;

  FILE *fp = fopen(filename,"rb");
  if (!fp)
  {
	  LOG(INFO)<<"Error: failed to load hash labels from "<<filename;
	  exit(0);
  }
  if (!loadHierHashes(m_vpHierHashes,fp))
  {
	  LOG(INFO)<<"Error: failed to load hashes from "<<filename;
	  fclose(fp);
	  exit(0);
  }
  if (!Load_labels(label_blob_,fp))
  {
	  LOG(INFO) << "Error: failed to load labels from "<<filename;
	  fclose(fp);
	  exit(0);
  }

  fclose(fp);

  CHECK_EQ((int)m_vpHierHashes.size(), label_blob_.shape(0)) << "hash num should == label num.";

  // Default to identity permutation.
  if (data_permutation_.size()!= m_vpHierHashes.size())
  {
	  data_permutation_.resize(m_vpHierHashes.size());
  }
  for (int i = 0; i < (int)m_vpHierHashes.size(); i++)
	  data_permutation_[i] = i;
 

  // Shuffle if needed.
  if (this->layer_param_.hash_data_param().shuffle()) {
	  std::random_shuffle(data_permutation_.begin(), data_permutation_.end());
	  DLOG(INFO) << "Successfully loaded " << (int)m_vpHierHashes.size()
		  << " hashes (shuffled)";
  }
  else {
	  DLOG(INFO) << "Successfully loaded " << (int)m_vpHierHashes.size() << " hashes";
  }

  //init the global structure information
  if (!CGlobalSys<Dtype>::isCreated())
  {
	  LOG(INFO) << "Fatal error: global data holder not inited!";
	  exit(0);
  }
  init_global_structure_blobs(CGlobalSys<Dtype>::get_struct_blobs());
}

template<typename Dtype>
void HashDataLayer<Dtype>::init_global_structure_blobs(std::vector<std::vector<Blob<int>*>> &structure_blobs)
{
	if (!m_vpHierHashes.size())
		return;
	const int struct_num = (int)m_vpHierHashes[0]->m_vpStructs.size();

	//already inited
	if ((int)structure_blobs.size() == struct_num)
		return;
	
	for (int i = 0; i < (int)structure_blobs.size(); i++)
	{
		for (int j = 0; j < (int)structure_blobs[i].size(); j++)
		{
			delete structure_blobs[i][j];
		}
	}

	structure_blobs.resize(struct_num);
	for (int i = 0; i < struct_num; i++)
	{
		structure_blobs[i].resize(HASH_STRUCTURE_SIZE);
		for (int j = 0; j < (int)structure_blobs[i].size(); j++)
		{
			structure_blobs[i][j] = new caffe::Blob<int>();
		}
	}
}


template <typename Dtype>
void HashDataLayer<Dtype>::BatchHierHashes_2_blobs(const std::vector<CHierarchyHash *> &batch_hashes, const std::vector<Blob<Dtype>*>& top_blobs,
	const std::vector<std::vector<Blob<int>*>> &structure_blobs)
{
	//check batch size
	const int batch_num = this->layer_param_.hash_data_param().batch_size();
	if (!batch_hashes.size()|| (int)batch_hashes.size()!=batch_num)
	{
		LOG(INFO)<<"Fatal error: BatchHierHashes_2_blobs failed! no hierHashes or batch size not match!";
		exit(0);
	}

	//check hash structure num and resize
	const int struct_num = (int)batch_hashes[0]->m_vpStructs.size();
	if (!struct_num || struct_num!=(int)structure_blobs.size())
	{
		LOG(INFO) << "Fatal error: BatchHierHashes_2_blobs failed! no structures or structure num not match! "<<struct_num<<" "<< (int)structure_blobs.size();
		exit(0);
	}

	const int channels = batch_hashes[0]->m_channels;
	if (channels<1)
	{
		LOG(INFO) << "Fatal error: BatchHierHashes_2_blobs failed! no channels!";
		exit(0);
	}

	//top blob only stores the valid data (Channels * all_valid_num size) and label
	if (top_blobs.size() != HASH_DATA_SIZE + 1)	//data + label
																					//structure info for multi layers + one hash data + one label
	{
		LOG(INFO) << "Error: hash data layer: top blobs should be have "<<HASH_DATA_SIZE<<" data and 1 label!";
		exit(0);
	}

	//reshape and set the structure blobs
	std::vector<int> batch_shape(1, batch_num);
	std::vector<int> batch_shape_1(1, batch_num + 1);
	//scalar for each hash in the batch
	for (int i = 0; i<struct_num; i++)
	{
		structure_blobs[i][M_BAR_BLOB]->Reshape(batch_shape);
		structure_blobs[i][R_BAR_BLOB]->Reshape(batch_shape);
		structure_blobs[i][DEFNUM_BLOB]->Reshape(batch_shape);
		structure_blobs[i][DEFNUM_SUM_BLOB]->Reshape(batch_shape_1);
		structure_blobs[i][M_SUM_BLOB]->Reshape(batch_shape_1);
		structure_blobs[i][R_SUM_BLOB]->Reshape(batch_shape_1);
		structure_blobs[i][RESOLUTION_BLOB]->Reshape(batch_shape);
	}

	//fill the value of m_bar, r_bar, define_num, and gather the total size for batched offset and pos_tag, 
	//gather the total size of m, and record each m && r, offset, postag, and bottom_hash_data
	for (int si = 0; si < struct_num; si++)
	{
		Blob<int>* hash_table_blob = structure_blobs[si][HASH_TABLE_BLOB];
		Blob<int>* offset_blob = structure_blobs[si][OFFSET_BLOB];
		Blob<int>* postag_blob = structure_blobs[si][POSTAG_BLOB];
		Blob<int>* volIdx_blob = structure_blobs[si][VOLUME_IDX_BLOB];

		Blob<int>* mBar_blob = structure_blobs[si][M_BAR_BLOB];
		Blob<int>* rBar_blob = structure_blobs[si][R_BAR_BLOB];
		Blob<int>* defNum_blob = structure_blobs[si][DEFNUM_BLOB];
		Blob<int>* defNumSum_blob = structure_blobs[si][DEFNUM_SUM_BLOB];
		Blob<int>* mSum_blob = structure_blobs[si][M_SUM_BLOB];
		Blob<int>* rSum_blob = structure_blobs[si][R_SUM_BLOB];
		Blob<int>* res_blob = structure_blobs[si][RESOLUTION_BLOB];

		//scalar for each hash in the batch
		int batch_m = 0;
		int batch_r = 0;
		int m_bar, r_bar, defNum, res;
		int m, r;	//m_bar*m_bar*m_bar, r_bar*r_bar*r_bar
		int totalDefNum = 0;
		for (int idx = 0; idx<batch_num; idx++)
		{
			m_bar = batch_hashes[idx]->m_vpStructs[si]->m_mBar;
			r_bar = batch_hashes[idx]->m_vpStructs[si]->m_rBar;
			defNum = batch_hashes[idx]->m_vpStructs[si]->m_validNum;
			res = batch_hashes[idx]->m_vpStructs[si]->m_res;

			mBar_blob->mutable_cpu_data()[idx] = m_bar;
			rBar_blob->mutable_cpu_data()[idx] = r_bar;
			defNum_blob->mutable_cpu_data()[idx] = defNum;
			res_blob->mutable_cpu_data()[idx] = res;

			if (!defNum)
			{
				LOG(INFO)<<"Fatal error: defNum zero! cur row: "<<current_row_<<"idx: "<<idx;
				batch_hashes[idx]->save("wrong_hierHash.hst");
				exit(0);
			}

			defNumSum_blob->mutable_cpu_data()[idx] = totalDefNum;
			mSum_blob->mutable_cpu_data()[idx] = batch_m;
			rSum_blob->mutable_cpu_data()[idx] = batch_r;

			batch_m += m_bar * m_bar * m_bar;
			batch_r += r_bar * r_bar * r_bar;
			totalDefNum += defNum;
		}
		defNumSum_blob->mutable_cpu_data()[batch_num] = totalDefNum;
		mSum_blob->mutable_cpu_data()[batch_num] = batch_m;
		rSum_blob->mutable_cpu_data()[batch_num] = batch_r;

		//if (si == 0)
		//{
		//	LOG(INFO) << "m sum: " << batch_m << "; " << "r sum: " << batch_r << "; " << "def sum: " << totalDefNum;
		//}

		//reshape the hash_index_table, offset_table, postag, and vol_idx
		//hash index table
		const int hash_table_size = batch_m;
		std::vector<int> hash_table_shape(1, hash_table_size);
		hash_table_blob->Reshape(hash_table_shape);
		//offset table
		const int offset_size = (batch_r * 3 * sizeof(unsigned char)) / sizeof(int) + 1;
		std::vector<int> offset_shape(1, offset_size);
		offset_blob->Reshape(offset_shape);
		//position tag
		const int posTag_size = (batch_m * sizeof(PACKED_POSITION)) / sizeof(int) + 1;
		std::vector<int> posTag_shape(1, posTag_size);
		postag_blob->Reshape(posTag_shape);
		//volume index of each hash table slot
		const int volIdx_size = (batch_m * sizeof(VolumeIndexType)) / sizeof(int) + 1;
		std::vector<int> volIdx_shape(1, volIdx_size);
		volIdx_blob->Reshape(volIdx_shape);

		//fill the hash_index_table, offset_table, postag, and vol_idx
		int* batch_hashtable_ptr = (int*)hash_table_blob->mutable_cpu_data();
		unsigned char *batch_offset_ptr = (unsigned char *)offset_blob->mutable_cpu_data();
		PACKED_POSITION *batch_posTag_ptr = (PACKED_POSITION *)postag_blob->mutable_cpu_data();
		VolumeIndexType* batch_volIdx_ptr = (VolumeIndexType*)volIdx_blob->mutable_cpu_data();

		for (int idx = 0; idx < batch_num; idx++)
		{
			m_bar = batch_hashes[idx]->m_vpStructs[si]->m_mBar;
			r_bar = batch_hashes[idx]->m_vpStructs[si]->m_rBar;
			m = m_bar*m_bar*m_bar;
			r = r_bar*r_bar*r_bar;

			memcpy(batch_hashtable_ptr, batch_hashes[idx]->m_vpStructs[si]->m_hash_index_table, sizeof(int)*m);
			memcpy(batch_offset_ptr, batch_hashes[idx]->m_vpStructs[si]->m_offset_data, sizeof(unsigned char)*r * 3);
			memcpy(batch_posTag_ptr, batch_hashes[idx]->m_vpStructs[si]->m_position_tag, sizeof(PACKED_POSITION)*m);
			for (int tmp_i = 0; tmp_i < m; tmp_i++)
				batch_volIdx_ptr[tmp_i] = (VolumeIndexType)idx;

			batch_offset_ptr += r * 3;
			batch_posTag_ptr += m;
			batch_hashtable_ptr += m;
			batch_volIdx_ptr += m;
		}


		//fill the bottom valid data
		if (si == 0)	//bottom hash
		{
			Blob<Dtype>* data_blob = top_blobs[HASH_DATA_BLOB];
			std::vector<int> data_shape;
#if ADD_NUM_2_SHAPE
			data_shape.push_back(1);
#endif
			data_shape.push_back(channels);
			data_shape.push_back(totalDefNum);
			data_blob->Reshape(data_shape);

			Dtype *data_ptr = data_blob->mutable_cpu_data();
			for (int idx = 0; idx < batch_num; idx++)
			{
				defNum = batch_hashes[idx]->m_vpStructs[si]->m_validNum;

				const float* cur_src_data = batch_hashes[idx]->m_valid_data;
				Dtype* cur_tgt_data = data_ptr;

				for (int c = 0; c < channels; c++)	//channel is row, defNu is col; row-major
				{
					for (int colIdx = 0; colIdx < defNum; colIdx++)
					{
						cur_tgt_data[colIdx] = (Dtype)(cur_src_data[colIdx]);
					}
					cur_src_data += defNum;
					cur_tgt_data += totalDefNum;
				}

				data_ptr += defNum;	//go to the next starting col for next hash in the batch
			}
		}
	}
}


template <typename Dtype>
void HashDataLayer<Dtype>::LayerSetUp(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) 
{
  // Refuse transformation parameters since HDF5 is totally generic.
  CHECK(!this->layer_param_.has_transform_param()) <<
      this->type() << " does not transform data.";
  // Read the source to parse the filenames.
  const string& source = this->layer_param_.hash_data_param().source();
  LOG(INFO) << "Loading list of HashLabel filenames from: " << source;

  segment_ = this->layer_param_.hash_data_param().segment();
  if(segment_)
	LOG(INFO) << "****This HashData is for segmentation***";

  hash_filenames_.clear();
  std::ifstream source_file(source.c_str());
  if (source_file.is_open()) {
    std::string line;
    while (source_file >> line) {
		hash_filenames_.push_back(line);
    }
  } else {
    LOG(FATAL) << "Failed to open source file: " << source;
  }
  source_file.close();
  num_files_ = hash_filenames_.size();
  current_file_ = 0;
  LOG(INFO) << "Number of HashLabel files: " << num_files_;
  CHECK_GE(num_files_, 1) << "Must have at least 1 HashLabel filename listed in "
    << source;

  file_permutation_.clear();
  file_permutation_.resize(num_files_);
  // Default to identity permutation.
  for (int i = 0; i < num_files_; i++) {
    file_permutation_[i] = i;
  }

  // Shuffle if needed.
  if (this->layer_param_.hash_data_param().shuffle()) {
    std::random_shuffle(file_permutation_.begin(), file_permutation_.end());
  }

  // Load the first HashLabel file and initialize the line counter.
  if (segment_)
  {
	  loadHashDataHashLabels(hash_filenames_[file_permutation_[current_file_]].c_str());
  }
  else
  {
	  LoadHashLabelFileData(hash_filenames_[file_permutation_[current_file_]].c_str());
  }
  
  current_row_ = 0;

  //init batch permutation
  const int batch_num = this->layer_param_.hash_data_param().batch_size();
  destroyBatch();
  m_hierHash_batch.resize(batch_num);
  for (int i = 0; i < batch_num; i++)
  {
	  m_hierHash_batch[i] = new CHierarchyHash();
	  *m_hierHash_batch[i] = *m_vpHierHashes[data_permutation_[i]];
  }


  if (segment_)
  {
	  destroyBatchHashLabels();
	  m_curBatchHashLabels.resize(batch_num);
	  for (int i = 0; i < batch_num; i++)
	  {
		  m_curBatchHashLabels[i] = new CHashLabels();
		  *m_curBatchHashLabels[i] = *m_vpHashLabels[data_permutation_[i]];
	  }
  }


  //send the first batch to tops in the setup stage, in order to allocate essential memories
  BatchHierHashes_2_blobs(m_hierHash_batch, top, CGlobalSys<Dtype>::get_struct_blobs());
  

  if (segment_)
  {
	  const int structure_num = m_vpHierHashes[0]->m_vpStructs.size();
	  int label_blob_idx = HASH_DATA_SIZE;
	  BatchHashLabels_2_blobs(m_curBatchHashLabels, top[label_blob_idx]);
  }
  else
  {
	  //reshape the top label, as the other blobs' shape will change every batch
	  vector<int> top_label_shape;
	  top_label_shape.resize(label_blob_.num_axes());
	  top_label_shape[0] = batch_num;
	  for (int j = 1; j < top_label_shape.size(); ++j)
	  {
		  top_label_shape[j] = label_blob_.shape(j);
	  }

	  int label_blob_idx = HASH_DATA_SIZE;
	  top[label_blob_idx]->Reshape(top_label_shape);
  }

#if MULT_THREAD
  StartInternalThread();
  LOG(INFO) << "Prefetch initialized.";
#endif
}

template <typename Dtype>
bool HashDataLayer<Dtype>::Skip() {
  int size = Caffe::solver_count();
  int rank = Caffe::solver_rank();
  bool keep = (offset_ % size) == rank ||
              // In test mode, only rank 0 runs, so avoid skipping
              this->layer_param_.phase() == TEST;
  return !keep;
}

template<typename Dtype>
void HashDataLayer<Dtype>::Next() 
{
  if (++current_row_ == (int)m_vpHierHashes.size()) {
    if (num_files_ > 1) {
      ++current_file_;
      if (current_file_ == num_files_) {
        current_file_ = 0;
        if (this->layer_param_.hash_data_param().shuffle()) {
          std::random_shuffle(file_permutation_.begin(),
                              file_permutation_.end());
        }
        DLOG(INFO) << "Looping around to first file.";
      }
	  if (segment_)
	  {
		  loadHashDataHashLabels(
			  hash_filenames_[file_permutation_[current_file_]].c_str());
	  }
	  else
	  {
		  LoadHashLabelFileData(
			  hash_filenames_[file_permutation_[current_file_]].c_str());
	  }
    }
    current_row_ = 0;
    if (this->layer_param_.hash_data_param().shuffle())
      std::random_shuffle(data_permutation_.begin(), data_permutation_.end());
  }
  offset_++;
}


template <typename Dtype>
void HashDataLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top) 
{
	const int batch_size = this->layer_param_.hash_data_param().batch_size();
	const int structure_num = m_vpHierHashes[0]->m_vpStructs.size();
	const int label_blob_idx = HASH_DATA_SIZE;
	const int label_dim = top[label_blob_idx]->count() / top[label_blob_idx]->shape(0);
	for (int i = 0; i < batch_size; ++i) {
		while (Skip()) {
#if MULT_THREAD
			Next_from_prefetch();
#else
			Next();
#endif
		}

		//record batch hash, will be send to top at last
		*m_hierHash_batch[i] = *m_vpHierHashes[data_permutation_[current_row_]];
		//send labels
		//printf("current row %d\n",current_row_);
		if (segment_)
		{
			*m_curBatchHashLabels[i] = *m_vpHashLabels[data_permutation_[current_row_]];
		}
		else
		{
			caffe_copy(label_dim,
				&label_blob_.cpu_data()[data_permutation_[current_row_]
				* label_dim], &top[label_blob_idx]->mutable_cpu_data()[i * label_dim]);
		}

#if MULT_THREAD
		Next_from_prefetch();
#else
		Next();
#endif
	}

	BatchHierHashes_2_blobs(m_hierHash_batch, top, CGlobalSys<Dtype>::get_struct_blobs());
	if (segment_)
	{
		BatchHashLabels_2_blobs(m_curBatchHashLabels, top[label_blob_idx]);
	}
}


template <typename Dtype>
void HashDataLayer<Dtype>::InternalThreadEntry() 
{
	try {
		while (!must_stop()) {		
			HierHash_Label_HashLabel_Batch<Dtype>* batch = prefetch_free_.pop();
			fetch_batch(batch);
			prefetch_full_.push(batch);
			//if (this->layer_param_.phase() == TEST)
			//{
			//	LOG(INFO) << "TEST full size: " << prefetch_full_.size();
			//}
			//else
			//{
			//	LOG(INFO) << "TRAIN full size: " << prefetch_full_.size();
			//}
		}
	}
	catch (boost::thread_interrupted&) {
		// Interrupted exception is expected on shutdown
	}
}

template <typename Dtype>
void HashDataLayer<Dtype>::fetch_batch(HierHash_Label_HashLabel_Batch<Dtype>* batch)
{
	if (num_files_ > 1) 
	{
		++current_file_;
		if (current_file_ == num_files_) 
		{
			current_file_ = 0;
			if (this->layer_param_.hash_data_param().shuffle()) {
				std::random_shuffle(file_permutation_.begin(),
					file_permutation_.end());
			}
			DLOG(INFO) << "Looping around to first file.";
		}
		if (segment_)
		{
			PreLoadHashDataHashLabels(batch->m_vpHierHashes, batch->m_vpHashLabels,
				hash_filenames_[file_permutation_[current_file_]].c_str());
		}
		else
		{
			PreLoadHashLabelFileData(batch->m_vpHierHashes, batch->m_label_blob,
				hash_filenames_[file_permutation_[current_file_]].c_str());
		}
	}
}

#ifdef CPU_ONLY
STUB_GPU_FORWARD(HashDataLayer, Forward);
#endif

template <typename Dtype>
void HashDataLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{
	Forward_cpu(bottom, top);
}

/****************************************for segmentation*****************************/
template <typename Dtype>
void HashDataLayer<Dtype>::destroyHashLabels()
{
	for (int i = 0; i<(int)m_vpHashLabels.size(); i++)
	{
		delete m_vpHashLabels[i];
	}
	m_vpHashLabels.resize(0);
}


// Load data and label from Hash filename into the class property blobs.
template <typename Dtype>
bool HashDataLayer<Dtype>::loadHashDataHashLabels(const char* filename)
{
	//NOTE: we dont destroy the memory to avoid the memory cost; We only enlarge the memory if needed
	//destroyHashes(m_hashes);
	LOG(INFO) << "Loading HashDataHashLabel file: " << filename;

	FILE *fp = fopen(filename, "rb");
	if (!fp)
	{
		printf("Error: failed to load hash labels from %s\n", filename);
		exit(0);
	}
	if (!loadHierHashes(m_vpHierHashes, fp))
	{
		printf("Error: failed to load hashes from %s\n", filename);
		fclose(fp);
		exit(0);
	}
	if (!load_hashLabels(m_vpHashLabels,fp))
	{
		printf("Error: failed to load labels from %s\n", filename);
		fclose(fp);
		exit(0);
	}

	fclose(fp);

	CHECK_EQ(m_vpHierHashes.size(), m_vpHashLabels.size()) << "hash num should == label num.";

	// Default to identity permutation.
	if (data_permutation_.size() != m_vpHierHashes.size())
	{
		data_permutation_.resize(m_vpHierHashes.size());
	}
	for (int i = 0; i < (int)m_vpHierHashes.size(); i++)
		data_permutation_[i] = i;


	// Shuffle if needed.
	if (this->layer_param_.hash_data_param().shuffle()) {
		std::random_shuffle(data_permutation_.begin(), data_permutation_.end());
		LOG(INFO) << "Successfully loaded " << m_vpHierHashes.size()
			<< " hashes (shuffled)";
	}
	else {
		LOG(INFO) << "Successfully loaded " << m_vpHierHashes.size() << " hashes";
	}

	//init the global structure information
	if (!CGlobalSys<Dtype>::isCreated())
	{
		LOG(INFO) << "Fatal error: global data holder not inited!";
		exit(0);
	}
	init_global_structure_blobs(CGlobalSys<Dtype>::get_struct_blobs());

	return true;
}


template <typename Dtype>
bool HashDataLayer<Dtype>::load_hashLabels(std::vector<CHashLabels*> &vpHashLabels, FILE *fp)
{
	int n;
	fread(&n, sizeof(int), 1, fp);

	if (n <= 0)
	{
		printf("Fatal error: hier hash num <= 0!\n");
		exit(0);
	}

	//actually when load batches, n will always to the same
	int ori_n = (int)vpHashLabels.size();
	if (ori_n<n)
	{
		for (int i = ori_n; i<n; i++)
		{
			vpHashLabels.push_back(new CHashLabels());
		}
	}
	if (ori_n>n)
	{
		for (int i = ori_n - 1; i >= n; i--)
		{
			delete vpHashLabels[i];
			vpHashLabels.pop_back();
		}
	}

	for (int i = 0; i < n; i++)
	{
		vpHashLabels[i]->load(fp);
	}
	return true;
}

template <typename Dtype>
void HashDataLayer<Dtype>::destroyBatchHashLabels()
{
	for (int i = 0; i<(int)m_curBatchHashLabels.size(); i++)
	{
		delete m_curBatchHashLabels[i];
	}
	m_curBatchHashLabels.resize(0);
}

template <typename Dtype>
void HashDataLayer<Dtype>::BatchHashLabels_2_blobs(const std::vector<CHashLabels *> &vpHashLabels, Blob<Dtype>* top_label)
{
	int total_m = 0;
	for (int i = 0; i < (int)vpHashLabels.size(); i++)
	{
		total_m += vpHashLabels[i]->m_m;
	}
	std::vector<int> label_shape;
#if ADD_NUM_2_SHAPE
	//in this case, total_m is the spatial dim
	label_shape.push_back(1);	//N
#endif
	label_shape.push_back(total_m);
	top_label->Reshape(label_shape);

	Dtype *top_ptr = top_label->mutable_cpu_data();
	for (int i = 0; i < (int)vpHashLabels.size(); i++)
	{
		for (int j = 0; j < vpHashLabels[i]->m_m; j++)
		{
			top_ptr[j] = (Dtype)(vpHashLabels[i]->m_pLabels[j]);
		}
		//memcpy(top_ptr, vpHashLabels[i]->m_pLabels,sizeof(float)*vpHashLabels[i]->m_m);
		top_ptr += vpHashLabels[i]->m_m;
	}

}


//new added for multi-thread

// Load data and label from Hash filename into the class property blobs.
template <typename Dtype>
void HashDataLayer<Dtype>::PreLoadHashLabelFileData(std::vector<CHierarchyHash*> &vpHierHashes, Blob<Dtype> &label_blob,const char* filename)
{
	if(this->layer_param_.phase() == TEST)
		LOG(INFO) << "TEST PreLoading HashLabel file: " << filename;
	else
		LOG(INFO) << "TRAIN PreLoading HashLabel file: " << filename;

	FILE *fp = fopen(filename, "rb");
	if (!fp)
	{
		printf("Error: failed to load hash labels from %s\n", filename);
		exit(0);
	}
	if (!loadHierHashes(vpHierHashes, fp))
	{
		printf("Error: failed to load hashes from %s\n", filename);
		fclose(fp);
		exit(0);
	}
	if (!Load_labels(label_blob, fp))
	{
		printf("Error: failed to load labels from %s\n", filename);
		fclose(fp);
		exit(0);
	}

	fclose(fp);

	CHECK_EQ((int)vpHierHashes.size(), label_blob.shape(0)) << "hash num should == label num.";
}

// Load data and label from Hash filename into the class property blobs.
template <typename Dtype>
bool HashDataLayer<Dtype>::PreLoadHashDataHashLabels(std::vector<CHierarchyHash*> &vpHierHashes, std::vector<CHashLabels*> &vpHashLabels,const char* filename)
{
	//NOTE: we dont destroy the memory to avoid the memory cost; We only enlarge the memory if needed
	if (this->layer_param_.phase() == TEST)
		LOG(INFO) << "TEST PreLoading HashDataHashLabel file: " << filename;
	else
		LOG(INFO) << "TRAIN PreLoading HashDataHashLabel file: " << filename;

	FILE *fp = fopen(filename, "rb");
	if (!fp)
	{
		printf("Error: failed to load hash labels from %s\n", filename);
		exit(0);
	}
	if (!loadHierHashes(vpHierHashes, fp))
	{
		printf("Error: failed to load hashes from %s\n", filename);
		fclose(fp);
		exit(0);
	}
	if (!load_hashLabels(vpHashLabels, fp))
	{
		printf("Error: failed to load labels from %s\n", filename);
		fclose(fp);
		exit(0);
	}

	fclose(fp);

	CHECK_EQ(vpHierHashes.size(), vpHashLabels.size()) << "hash num should == label num.";

	return true;
}


template <typename Dtype>
void HashDataLayer<Dtype>::update_from_preload(HierHash_Label_HashLabel_Batch<Dtype>* prefetch)
{
	//simply swap the pointers between member and prefetch
	std::vector<CHierarchyHash*> vpTmpHierHashes(m_vpHierHashes.begin(),m_vpHierHashes.end());
	m_vpHierHashes.assign(prefetch->m_vpHierHashes.begin(), prefetch->m_vpHierHashes.end());
	prefetch->m_vpHierHashes.assign(vpTmpHierHashes.begin(), vpTmpHierHashes.end());

	if (segment_)
	{
		//simply swap pointers
		std::vector<CHashLabels*> vpTmpHashLabels(m_vpHashLabels.begin(), m_vpHashLabels.end());
		m_vpHashLabels.assign(prefetch->m_vpHashLabels.begin(), prefetch->m_vpHashLabels.end());
		prefetch->m_vpHashLabels.assign(vpTmpHashLabels.begin(), vpTmpHashLabels.end());
	}
	else
	{
		label_blob_.ReshapeLike(prefetch->m_label_blob);
		caffe::caffe_copy(label_blob_.count(), prefetch->m_label_blob.cpu_data(), label_blob_.mutable_cpu_data());
	}


	if (data_permutation_.size() != m_vpHierHashes.size())
	{
		data_permutation_.resize(m_vpHierHashes.size());
	}
	for (int i = 0; i < (int)m_vpHierHashes.size(); i++)
		data_permutation_[i] = i;

	//// Shuffle if needed.
	//if (this->layer_param_.hash_data_param().shuffle()) 
	//{
	//	std::random_shuffle(data_permutation_.begin(), data_permutation_.end());
	//	DLOG(INFO) << "Successfully loaded " << m_vpHierHashes.size()
	//		<< " hashes (shuffled)";
	//}
	//else {
	//	DLOG(INFO) << "Successfully loaded " << m_vpHierHashes.size() << " hashes";
	//}
}

template<typename Dtype>
void HashDataLayer<Dtype>::Next_from_prefetch()
{
	if (++current_row_ == (int)m_vpHierHashes.size()) 
	{
		if (num_files_ > 1) 
		{
			if (prefetch_current_) {
				prefetch_free_.push(prefetch_current_);
			}
			prefetch_current_ = prefetch_full_.pop("Waiting for data");
			update_from_preload(prefetch_current_);
		}
		current_row_ = 0;
		if (this->layer_param_.hash_data_param().shuffle())
			std::random_shuffle(data_permutation_.begin(), data_permutation_.end());
	}
	offset_++;
}

INSTANTIATE_CLASS(HashDataLayer);
REGISTER_LAYER_CLASS(HashData);
}  // namespace caffe
