#include <algorithm>
#include <cfloat>
#include <vector>


#include "caffe/layer.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/layers/pool_hash_layer.hpp"
#include "caffe\util\HashData.h"

namespace caffe {

	template <typename Dtype>
	__global__ void forward_gpu_max_kernel(const Dtype* bottom_data, 
		const CBatchStructure bottom_struct,
		Dtype *top_data, int *mask, Dtype *top_mask,
		const CBatchStructure top_struct,
		int top_total_m, int channels, 
		int pad, int stride, int kernel_size
		)
	{
		//nThreads = channels * kernel_size * top_total_m
		CUDA_KERNEL_LOOP(cur_thread, channels * top_total_m)
		{
			const int bottom_total_defNum = bottom_struct.m_defNumSum_ptr[bottom_struct.m_batch_size];
			const int top_total_defNum = top_struct.m_defNumSum_ptr[top_struct.m_batch_size];

			int C = cur_thread / top_total_m;
			int top_M_idx = cur_thread - C*top_total_m;

			const int top_valid_idx = top_struct.m_hashIdxTable_ptr[top_M_idx];	//valid idx in the hash within the batch
			const VolumeIndexType volIdx = top_struct.m_volIdx_ptr[top_M_idx];	//the idx of the hash in the batch
			const int top_defNumSum = (int)top_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash

			const int bt_dense_res = (int)bottom_struct.m_res_ptr[volIdx];	//dense resolution
			const int bt_m_bar = (int)bottom_struct.m_mBar_ptr[volIdx];	//m bar
			const int bt_r_bar = (int)bottom_struct.m_rBar_ptr[volIdx];	//r bar
			const int bt_rSum = (int)bottom_struct.m_rSum_ptr[volIdx];	//rSum before this hash
			const int bt_mSum = (int)bottom_struct.m_mSum_ptr[volIdx];	//rSum before this hash
			const int bt_defNumSum = (int)bottom_struct.m_defNumSum_ptr[volIdx];
			const unsigned char* bt_offset = bottom_struct.m_offset_ptr + bt_rSum * 3;
			const PACKED_POSITION *bt_posTags = bottom_struct.m_posTag_ptr + bt_mSum;
			const int *bt_hash_index_table = bottom_struct.m_hashIdxTable_ptr + bt_mSum;
			if (top_valid_idx < 0)
			{
				return;	//invalid hash slot, empty
			}

			const int3 top_voxel = xyz_from_pack_g(top_struct.m_posTag_ptr[top_M_idx]);

			const int global_top_valid_idx = top_defNumSum + top_valid_idx;	
			const int top_data_idx = C*top_total_defNum + global_top_valid_idx;

			//init to min
			top_data[top_data_idx] = -FLT_MAX;

			int min_x = top_voxel.x * stride - pad;
			int min_y = top_voxel.y * stride - pad;
			int min_z = top_voxel.z * stride - pad;
			int x_end = min(min_x + kernel_size, bt_dense_res);
			int y_end = min(min_y + kernel_size, bt_dense_res);
			int z_end = min(min_z + kernel_size, bt_dense_res);
			min_x = max(min_x, 0);
			min_y = max(min_y, 0);
			min_z = max(min_z, 0);

			for (int nz = min_z; nz < z_end; ++nz)
			{ 
				for (int ny = min_y; ny < y_end; ++ny)
				{
					for (int nx = min_x; nx < x_end; ++nx)
					{
						//hash to get hash position
						const int bt_m_idx = NXYZ2I_g(Hash_g(nx, ny, nz, bt_offset,
							bt_m_bar, bt_r_bar), bt_m_bar);

						const int bt_valid_idx = bt_hash_index_table[bt_m_idx];
						if (bt_valid_idx < 0)	//invalid slot
							continue;
						const int3 stored = xyz_from_pack_g(bt_posTags[bt_m_idx]);
						//the bottom hash voxel is undefined
						if (nx != stored.x || ny != stored.y || nz != stored.z)
							continue;

						const int global_bt_valid_idx = bt_defNumSum + bt_valid_idx;
						const int bt_data_idx = C*bottom_total_defNum + global_bt_valid_idx;
						const Dtype bt_val = bottom_data[bt_data_idx];
						if (top_data[top_data_idx] < bt_val)
						{
							top_data[top_data_idx] = bt_val;
							if (mask)
								mask[top_data_idx] = global_bt_valid_idx;
							else
								top_mask[top_data_idx] = (Dtype)(global_bt_valid_idx);
						}
					}
				}
			} // end for nx, ny, nz
#if USE_EMPTY_VALID_REGION
			if (top_data[top_data_idx] == -FLT_MAX)//expanded regions, no parent
			{
				top_data[top_data_idx] = 0;
				if (mask)
					mask[top_data_idx] = -1;
				else
					top_mask[top_data_idx] = (Dtype)-1;
			}
#endif
		} // end for caffe_kernel_loop
	}


	template<class Dtype>
	__global__ void backward_gpu_max_kernel(
		Dtype *bottom_dif, int bottom_total_def_num,
		const Dtype *top_dif, int top_total_def_num,
		const int *mask, const Dtype *top_mask,
		const int channels)
	{
		CUDA_KERNEL_LOOP(threadId, top_total_def_num*channels)
		{
			const int c = threadId / top_total_def_num;
			const int global_bt_valid_idx = mask?mask[threadId]:(int)top_mask[threadId];
#if USE_EMPTY_VALID_REGION
			if(global_bt_valid_idx !=-1)
				bottom_dif[global_bt_valid_idx + c*bottom_total_def_num] = top_dif[threadId];
#else
			bottom_dif[global_bt_valid_idx + c*bottom_total_def_num] = top_dif[threadId];
#endif
		}
	}


	template <typename Dtype>
	void PoolHashLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
		const vector<Blob<Dtype>*>& top)
	{
		switch (this->layer_param_.pooling_param().pool()) {
		case PoolingParameter_PoolMethod_MAX:
			Forward_gpu_max(bottom, top);
			break;
		case PoolingParameter_PoolMethod_AVE:
			NOT_IMPLEMENTED;
			break;
		case PoolingParameter_PoolMethod_STOCHASTIC:
			NOT_IMPLEMENTED;
			break;
		default:
			LOG(FATAL) << "Unknown pooling method.";
		}
	}

	template <typename Dtype>
	void PoolHashLayer<Dtype>::Forward_gpu_max(const vector<Blob<Dtype>*>& bottom,
		const vector<Blob<Dtype>*>& top)
	{
		//struct GPU
		CBatchStructure bottom_struct_GPU;
		CGlobalSys<Dtype>::setBatchStructure_GPU(bottom_struct_GPU, m_bottom_struct_idx);	//top is the same as bottom
																							//get top struct
		CBatchStructure top_struct_GPU;
		CGlobalSys<Dtype>::setBatchStructure_GPU(top_struct_GPU, m_top_struct_idx);	//top is the same as bottom

																					//struct CPU
		CBatchStructure top_struct;
		CGlobalSys<Dtype>::setBatchStructure(top_struct, m_top_struct_idx);	//top is the same as bottom
		int top_total_m = (int)top_struct.m_mSum_ptr[top_struct.m_batch_size];

		bool use_top_mask = (top.size() == (HASH_DATA_SIZE + 1));
		int* mask = NULL;  // suppress warnings about uninitalized variables
		Dtype* top_mask = NULL;
		if (use_top_mask)
		{
			top_mask = top[HASH_DATA_SIZE]->mutable_gpu_data();
		}
		else
		{
			mask = max_idx_.mutable_gpu_data();
		}

		this->forward_gpu_max(bottom[HASH_DATA_BLOB]->gpu_data(), &bottom_struct_GPU,
			top[HASH_DATA_BLOB]->mutable_gpu_data(), &top_struct_GPU,
			mask, top_mask, top_total_m);


		//FILE *fp = fopen("max_idx_GPU.bin", "wb");
		//if (use_top_mask)
		//{
		//	fwrite(top[HASH_DATA_SIZE]->cpu_data(), sizeof(Dtype), top[HASH_DATA_SIZE]->count(), fp);
		//}
		//else
		//{
		//	fwrite(max_idx_.cpu_data(), sizeof(int), max_idx_.count(), fp);
		//}
		//fclose(fp);

		//for debug
		//if (top_struct.m_res_ptr[0] == 256)
		//{
		//	static int counter;
		//	char buf[128];
		//	sprintf(buf, "debug_pool_forward_%02d.txt",counter++);
		//	std::ofstream debug_ofs(buf);
		//	debug_ofs << "===============Top data==========================" << std::endl;
		//	const Dtype *top_data = top[HASH_DATA_BLOB]->cpu_data();
		//	for (int i = 0; i < top[HASH_DATA_BLOB]->shape(SPATIAL_AXIS); i++)
		//	{
		//		const Dtype *cur_data = top_data;
		//		for (int j = 0; j < top[HASH_DATA_BLOB]->shape(CHANNEL_AXIS); j++)
		//		{
		//			debug_ofs << (int)*cur_data << " ";
		//			cur_data += top[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
		//		}
		//		debug_ofs << std::endl;
		//		top_data++;
		//	}
		//	debug_ofs.close();
		//}
	}


	template <typename Dtype>
	void PoolHashLayer<Dtype>::forward_gpu_max(const Dtype *bottom_valid_data,
		const CBatchStructure *bottom_struct_GPU,
		Dtype *top_valid_data, const CBatchStructure *top_struct_GPU,
		int *mask, Dtype *top_mask, int top_total_m)
	{
		int nThreads = channels_ * top_total_m;
		const int stride = stride_shape_.cpu_data()[0];
		const int pad = pad_shape_.cpu_data()[0];
		const int kernel = kernel_shape_.cpu_data()[0];
		forward_gpu_max_kernel<Dtype> << <CAFFE_GET_BLOCKS(nThreads), CAFFE_CUDA_NUM_THREADS >> > (
			bottom_valid_data, *bottom_struct_GPU,
			top_valid_data, mask, top_mask, *top_struct_GPU,
			top_total_m, channels_, pad, stride, kernel);
	}

	template <typename Dtype>
	void PoolHashLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
		const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom)
	{
		if (!propagate_down[HASH_DATA_BLOB]) {
			return;
		}
		switch (this->layer_param_.pooling_param().pool()) {
		case PoolingParameter_PoolMethod_MAX:
			Backward_gpu_max(top, propagate_down, bottom);
			break;
		case PoolingParameter_PoolMethod_AVE:
			NOT_IMPLEMENTED;
			break;
		case PoolingParameter_PoolMethod_STOCHASTIC:
			NOT_IMPLEMENTED;
			break;
		default:
			LOG(FATAL) << "Unknown pooling method.";
		}
	}

	template <typename Dtype>
	void PoolHashLayer<Dtype>::Backward_gpu_max(const vector<Blob<Dtype>*>& top,
		const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom)
	{
		//struct GPU
		CBatchStructure bottom_struct_CPU;
		CGlobalSys<Dtype>::setBatchStructure(bottom_struct_CPU, m_bottom_struct_idx);	//top is the same as bottom
		const int bottom_total_def_num = bottom_struct_CPU.m_defNumSum_ptr[bottom_struct_CPU.m_batch_size];
																							//get top struct
		CBatchStructure top_struct_CPU;
		CGlobalSys<Dtype>::setBatchStructure(top_struct_CPU, m_top_struct_idx);	//top is the same as bottom
		const int top_total_def_num = top_struct_CPU.m_defNumSum_ptr[top_struct_CPU.m_batch_size];
																					//struct CPU
		
		bool use_top_mask = (top.size() == (HASH_DATA_SIZE + 1));
		const int* mask = NULL;  // suppress warnings about uninitalized variables
		const Dtype* top_mask = NULL;
		if (use_top_mask)
		{
			top_mask = top[HASH_DATA_SIZE]->gpu_data();
		}
		else
		{
			mask = max_idx_.gpu_data();
		}

		//init bottom dif to zero
		caffe_gpu_set(bottom[HASH_DATA_BLOB]->count(), Dtype(0.), bottom[HASH_DATA_BLOB]->mutable_gpu_diff());

		backward_gpu_max_kernel<Dtype> << <CAFFE_GET_BLOCKS(channels_*top_total_def_num), CAFFE_CUDA_NUM_THREADS >> > (
			bottom[HASH_DATA_BLOB]->mutable_gpu_diff(), bottom_total_def_num,
			top[HASH_DATA_BLOB]->gpu_diff(), top_total_def_num,
			mask, top_mask,channels_);
	}

	INSTANTIATE_LAYER_GPU_FUNCS(PoolHashLayer);
}  // namespace caffe
