#include <algorithm>
#include <vector>

#include "caffe/filler.hpp"
#include "caffe/layer_factory.hpp"
#include "caffe/layers/scale_hash_layer.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/util/MyMacro.h"
#include "caffe/util/HashData.h"
namespace caffe {

template <typename Dtype>
void ScaleHashLayer<Dtype>::LayerSetUp(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) {

	//We do not allow bottom and top share the same memory in this implementation
	if (bottom[HASH_DATA_BLOB] == top[HASH_DATA_BLOB])
	{
		LOG(INFO) << "Fatal error: We do not allow bottom and top share the same memory in this implementation!";
		exit(0);
	}

	const ScaleHashParameter& scale_hash_param = this->layer_param_.scale_hash_param();

	if (scale_hash_param.bias_term()) {
	  this->blobs_.resize(2);
	}
	else {
		this->blobs_.resize(1);
	}

	channels_ = (int)bottom[HASH_DATA_BLOB]->shape(CHANNEL_AXIS);
	CHECK_GT(channels_, 0);

	const ScaleHashParameter& param = this->layer_param_.scale_hash_param();
	// scale is a learned parameter; initialize it
	vector<int> scale_shape(1, channels_);
	this->blobs_[0].reset(new Blob<Dtype>(scale_shape));
	FillerParameter filler_param(param.filler());
	if (!param.has_filler()) {
		// Default to unit (1) filler for identity operation.
		filler_param.set_type("constant");
		filler_param.set_value(1);
	}
	shared_ptr<Filler<Dtype> > filler(GetFiller<Dtype>(filler_param));
	filler->Fill(this->blobs_[0].get());

	if (param.bias_term())
	{
		this->blobs_[1].reset(new Blob<Dtype>(scale_shape));
		FillerParameter filler_param(param.bias_filler());
		if (!param.has_bias_filler()) {
			// Default to unit (1) filler for identity operation.
			filler_param.set_type("constant");
			filler_param.set_value(0);
		}
		shared_ptr<Filler<Dtype> > filler(GetFiller<Dtype>(filler_param));
		filler->Fill(this->blobs_[1].get());
	}
	this->param_propagate_down_.resize(this->blobs_.size(), true);
}


template <typename Dtype>
void ScaleHashLayer<Dtype>::Reshape(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) 
{
	top[HASH_DATA_BLOB]->ReshapeLike(*bottom[HASH_DATA_BLOB]);

	Blob<Dtype> &global_buf = CGlobalSys<Dtype>::get_workspace();
	global_buf.ReshapeLike(*bottom[HASH_DATA_BLOB]);

	// a multipler for Mv
	Blob<Dtype> &multiplier = CGlobalSys<Dtype>::get_multiplier();
	vector<int> sz(1);
	sz[0] = bottom[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
	multiplier.Reshape(sz);
	switch (Caffe::mode())
	{
	default:
		break;
	case Caffe::GPU:
		caffe_gpu_set(multiplier.count(), Dtype(1), multiplier.mutable_gpu_data());
		break;
	case Caffe::CPU:
		caffe_set(multiplier.count(), Dtype(1), multiplier.mutable_cpu_data());
		break;
	}
}

template <typename Dtype>
void ScaleHashLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom, const vector<Blob<Dtype>*>& top) 
{
	Dtype* scale_data = this->blobs_[0]->mutable_cpu_data();
	if (this->layer_param_.scale_hash_param().has_min_value()) 
	{
		for (int d = 0; d < channels_; d++) {
			scale_data[d] = std::max<Dtype>(scale_data[d], this->layer_param_.scale_hash_param().min_value());
		}
	}
	if (this->layer_param_.scale_hash_param().has_max_value()) {
		for (int d = 0; d < channels_; d++) {
			scale_data[d] = std::min<Dtype>(scale_data[d], this->layer_param_.scale_hash_param().max_value());
		}
	}

	/***************************************************************************************/
	const Dtype *bias_data = this->blobs_[1]->cpu_data();
	Dtype* top_data = top[HASH_DATA_BLOB]->mutable_cpu_data();	
	const Dtype* bottom_data = bottom[HASH_DATA_BLOB]->cpu_data();
	int spatial_dim = top[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
	for (int c = 0; c < channels_; ++c) {
		caffe_cpu_scale(spatial_dim, scale_data[c], bottom_data, top_data);
		if (this->layer_param_.scale_hash_param().bias_term())
		{
			caffe_add_scalar(spatial_dim, bias_data[c], top_data);
		}
		bottom_data += spatial_dim;
		top_data += spatial_dim;
	}
	/***************************************************************************************/	
}


template <typename Dtype>
void ScaleHashLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
    const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) 
{
	const int total_def_num = top[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
	const Dtype *top_dif = top[HASH_DATA_BLOB]->cpu_diff();
	Dtype *bottom_dif = bottom[HASH_DATA_BLOB]->mutable_cpu_diff();
	const Dtype* bottom_data = bottom[HASH_DATA_BLOB]->cpu_data();
	const Blob<Dtype> &multiplier = CGlobalSys<Dtype>::get_multiplier();
	//bias bp
  if (this->layer_param_.scale_hash_param().bias_term() &&
      this->param_propagate_down_.back()) 
  {
	  caffe_cpu_gemv<Dtype>(CblasNoTrans, channels_, total_def_num, 1.,
		  top_dif, multiplier.cpu_data(), 0, this->blobs_[1]->mutable_cpu_diff());
  }
  //weight bp
  if (this->param_propagate_down_[0])
  {
	  // Hack: store big eltwise product in bottom[0] diff, except in the special
	  // case where this layer itself does the eltwise product, in which case we
	  // can store it directly in the scale diff, and we're done.
	  // If we're computing in-place (and not doing eltwise computation), this
	  // hack doesn't work and we store the product in temp_.
	  Dtype* product = bottom_dif;
	  caffe_mul(bottom[HASH_DATA_BLOB]->count(), top_dif, bottom_data, product);

	  caffe_cpu_gemv(CblasNoTrans, channels_, total_def_num,
		  Dtype(1), product, multiplier.cpu_data(), Dtype(0),
		  this->blobs_[0]->mutable_cpu_diff());
  }
 
   
  if (propagate_down[0])	
  {
    const Dtype* scale_data = this->blobs_[0]->cpu_data();
	top_dif = top[HASH_DATA_BLOB]->cpu_diff();
	bottom_dif = bottom[HASH_DATA_BLOB]->mutable_cpu_diff();
    for (int d = 0; d < channels_; ++d) 
	{
		const Dtype factor = scale_data[d];
		caffe_cpu_scale(total_def_num, factor, top_dif, bottom_dif);
		bottom_dif += total_def_num;
		top_dif += total_def_num;
    }
  }
}


#ifdef CPU_ONLY
STUB_GPU(ScaleHashLayer);
#endif

INSTANTIATE_CLASS(ScaleHashLayer);
REGISTER_LAYER_CLASS(ScaleHash);

}  // namespace caffe
