#include <algorithm>
#include <cfloat>
#include <vector>

#include "caffe/common.hpp"
#include "caffe/layer.hpp"
#include "caffe/syncedmem.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/layers/unpool_hash_layer.hpp"

namespace caffe {

using std::min;
using std::max;

template <typename Dtype>
void UnPoolHashLayer<Dtype>::LayerSetUp(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top) 
{
	//CHECK size
	if (bottom.size() != HASH_DATA_SIZE + 1)	//data + mask
	{
		printf("Fatal error: bottom size should be %d\n", HASH_DATA_SIZE + 1);
		exit(0);
	}
	if (top.size() != HASH_DATA_SIZE)
	{
		printf("Fatal error: top size should be %d\n",HASH_DATA_SIZE);
		exit(0);
	}


	num_spatial_axes_ = 3;	//for 3D case

	channels_ = bottom[HASH_DATA_BLOB]->shape(CHANNEL_AXIS);

	// Configure the kernel size, padding, stride, and inputs.
	const PoolHashParameter &pool_hash_param = this->layer_param_.pool_hash_param();
	m_bottom_struct_idx = pool_hash_param.bottom_struct();
	m_top_struct_idx = pool_hash_param.top_struct();

	const vector<int> spatial_dim_blob_shape(1, num_spatial_axes_);
	
	// Setup filter kernel dimensions(kernel_shape_).
	kernel_shape_.Reshape(spatial_dim_blob_shape);
	int* kernel_shape_data = kernel_shape_.mutable_cpu_data();
	{
		const int num_kernel_dims = pool_hash_param.kernel_size_size();
		CHECK(num_kernel_dims == 1 || num_kernel_dims == num_spatial_axes_)
			<< "kernel_size must be specified once, or once per spatial dimension "
			<< "(kernel_size specified " << num_kernel_dims << " times; "
			<< num_spatial_axes_ << " spatial dims).";
		for (int i = 0; i < num_spatial_axes_; ++i) {
			kernel_shape_data[i] =
				pool_hash_param.kernel_size((num_kernel_dims == 1) ? 0 : i);
		}
	}
	for (int i = 0; i < num_spatial_axes_; ++i) {
		CHECK_GT(kernel_shape_data[i], 0) << "Filter dimensions must be nonzero.";
	}

	// Setup filter pad dimensions(pad_shape_).
	pad_shape_.Reshape(spatial_dim_blob_shape);
	for (int i = 0; i < num_spatial_axes_; i++)
		pad_shape_.mutable_cpu_data()[i] = (pool_hash_param.pad_size() == 0 ? 0 :
			pool_hash_param.pad((pool_hash_param.pad_size() == 1) ? 0 : i));


	// Setup filter stride dimensions(stride_shape_).
	stride_shape_.Reshape(spatial_dim_blob_shape);
	for (int i = 0; i < num_spatial_axes_; i++)
		stride_shape_.mutable_cpu_data()[i] = (pool_hash_param.stride_size() == 0 ? 1 :
			pool_hash_param.stride((pool_hash_param.stride_size() == 1) ? 0 : i));

	//Currently we only support kernel_size==stride, because in pooling_backward, we avoid atomicAdd to fairly compare speed with OCNN
	if (stride_shape_.cpu_data()[0] != kernel_shape_.cpu_data()[0])
	{
		LOG(INFO) << "Currently we only support kernel_size==stride, because in pooling_backward, we avoid atomicAdd to fairly compare speed with OCNN";
		exit(0);
	}
}


template <typename Dtype>
void UnPoolHashLayer<Dtype>::Reshape(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{
	const Blob<int>* top_defNumSum_blob = CGlobalSys<Dtype>::get_defNumSum_blob(m_top_struct_idx);
	const int total_top_defNum = top_defNumSum_blob->cpu_data()[top_defNumSum_blob->count() - 1];
	std::vector<int> data_shape;
#if ADD_NUM_2_SHAPE
	data_shape.push_back(1);
#endif
	data_shape.push_back(channels_);
	data_shape.push_back(total_top_defNum);
	top[HASH_DATA_BLOB]->Reshape(data_shape);
}


template <typename Dtype>
void UnPoolHashLayer<Dtype>::forward_cpu_max(const Dtype *bottom_valid_data,
	const CBatchStructure *bottom_struct,
	Dtype *top_valid_data, const CBatchStructure *top_struct,
	const Dtype *mask)
{
	const int top_total_def_num = top_struct->m_defNumSum_ptr[top_struct->m_batch_size];
	const int bottom_total_def_num = bottom_struct->m_defNumSum_ptr[bottom_struct->m_batch_size];
	//init top data to zero
	caffe_set(top_total_def_num*channels_, (Dtype)0, top_valid_data);
	for (int v = 0; v < bottom_total_def_num; v++)
	{
		const Dtype *bt_data_ptr = &bottom_valid_data[v];
		const Dtype *mask_ptr = &mask[v];
		Dtype *tp_data_start = top_valid_data;

#if USE_EMPTY_VALID_REGION
		if (*mask_ptr == -1 )	//expanded regions, no parent
		{

		}
		else
		{
			for (int c = 0; c < channels_; c++)
			{
				const int tp_idx = *mask_ptr;

				tp_data_start[tp_idx] = *bt_data_ptr;

				tp_data_start += top_total_def_num;
				bt_data_ptr += bottom_total_def_num;
				mask_ptr += bottom_total_def_num;
			}
		}
#else
		for (int c = 0; c < channels_; c++)
		{
			const int tp_idx = *mask_ptr;

			tp_data_start[tp_idx] = *bt_data_ptr;

			tp_data_start += top_total_def_num;
			bt_data_ptr += bottom_total_def_num;
			mask_ptr += bottom_total_def_num;
		}
#endif
	}
}

template <typename Dtype>
void UnPoolHashLayer<Dtype>::Forward_cpu_max(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{
	//get bottom struct
	CBatchStructure bottom_struct;
	CGlobalSys<Dtype>::setBatchStructure(bottom_struct, m_bottom_struct_idx);	//top is the same as bottom
																				//get top struct
	CBatchStructure top_struct;
	CGlobalSys<Dtype>::setBatchStructure(top_struct, m_top_struct_idx);	//top is the same as bottom


	const Dtype* mask = bottom[HASH_DATA_SIZE]->cpu_data();

	forward_cpu_max(bottom[HASH_DATA_BLOB]->cpu_data(),
		&bottom_struct,
		top[HASH_DATA_BLOB]->mutable_cpu_data(), &top_struct, mask);
}


template <typename Dtype>
void UnPoolHashLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top) 
{
	switch (this->layer_param_.pooling_param().pool()) {
	case PoolingParameter_PoolMethod_MAX:
		Forward_cpu_max(bottom, top);
		break;
	case PoolingParameter_PoolMethod_AVE:
		NOT_IMPLEMENTED;
		break;
	case PoolingParameter_PoolMethod_STOCHASTIC:
		NOT_IMPLEMENTED;
		break;
	default:
		LOG(FATAL) << "Unknown pooling method.";
	}
}


template <typename Dtype>
void UnPoolHashLayer<Dtype>::backward_cpu_max(Dtype *bottom_valid_diff,
	const CBatchStructure *bottom_struct,
	const Dtype *top_valid_diff, const CBatchStructure *top_struct,
	const Dtype *mask)
{
	const int top_total_def_num = top_struct->m_defNumSum_ptr[top_struct->m_batch_size];
	const int bottom_total_def_num = bottom_struct->m_defNumSum_ptr[bottom_struct->m_batch_size];
	//init bottom diff to zero
	caffe_set(bottom_total_def_num*channels_, (Dtype)0, bottom_valid_diff);
	for (int v = 0; v < bottom_total_def_num; v++)
	{
		Dtype *bt_diff_ptr = &bottom_valid_diff[v];
		const Dtype *mask_ptr = &mask[v];
		const Dtype *tp_diff_start = top_valid_diff;

#if USE_EMPTY_VALID_REGION
		if (*mask_ptr == -1)	//expanded regions, no parent
		{

		}
		else
		{
			for (int c = 0; c < channels_; c++)
			{
				const int tp_idx = *mask_ptr;

				*bt_diff_ptr = tp_diff_start[tp_idx];

				tp_diff_start += top_total_def_num;
				bt_diff_ptr += bottom_total_def_num;
				mask_ptr += bottom_total_def_num;
			}
		}
#else
		for (int c = 0; c < channels_; c++)
		{
			const int tp_idx = *mask_ptr;

			*bt_diff_ptr = tp_diff_start[tp_idx];

			tp_diff_start += top_total_def_num;
			bt_diff_ptr += bottom_total_def_num;
			mask_ptr += bottom_total_def_num;
		}
#endif
	}
}

template <typename Dtype>
void UnPoolHashLayer<Dtype>::Backward_cpu_max(const vector<Blob<Dtype>*>& top,
	const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) 
{
	//get bottom struct
	CBatchStructure bottom_struct;
	CGlobalSys<Dtype>::setBatchStructure(bottom_struct, m_bottom_struct_idx);	//top is the same as bottom
																				//get top struct
	CBatchStructure top_struct;
	CGlobalSys<Dtype>::setBatchStructure(top_struct, m_top_struct_idx);	//top is the same as bottom

	const Dtype* mask = bottom[HASH_DATA_SIZE]->cpu_data();

	backward_cpu_max(bottom[HASH_DATA_BLOB]->mutable_cpu_diff(),
		&bottom_struct,
		top[HASH_DATA_BLOB]->cpu_diff(), &top_struct, mask);
}


template <typename Dtype>
void UnPoolHashLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
	const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) 
{
	if (!propagate_down[HASH_DATA_BLOB]) {
		return;
	}

	switch (this->layer_param_.pooling_param().pool()) {
	case PoolingParameter_PoolMethod_MAX:
		Backward_cpu_max(top, propagate_down,bottom);
		break;
	case PoolingParameter_PoolMethod_AVE:
		NOT_IMPLEMENTED;
		break;
	case PoolingParameter_PoolMethod_STOCHASTIC:
		NOT_IMPLEMENTED;
		break;
	default:
		LOG(FATAL) << "Unknown pooling method.";
	}
	//printf("******************UnPoolHash backward CPU end\n");
}



#ifdef CPU_ONLY
STUB_GPU(UnPoolHashLayer);
#endif

INSTANTIATE_CLASS(UnPoolHashLayer);
REGISTER_LAYER_CLASS(UnPoolHash);
}  // namespace caffe

