#include <vector>

#include "caffe/util/hash2col.hpp"
#include "caffe/util/math_functions.hpp"

namespace caffe {
	//NOTE: the matrix is row-major
	template<typename Dtype>
	int hash2col_self_cpu(const Dtype* valid_data, int channels, const CBatchStructure *self_struct,
		const int kernel_shape[3],	//D, H, W
		Dtype* col_buff)
	{
		const int batch_size = self_struct->m_batch_size;
		const int hx = (kernel_shape[0] >> 1);
		const int hy = (kernel_shape[1] >> 1);
		const int hz = (kernel_shape[2] >> 1);
		//col is reception field: input_channels * KD * KH * KW; row: spatial domain
		const int kernel_dim = kernel_shape[0] * kernel_shape[1] * kernel_shape[2];
		const int rows = kernel_dim * channels;
		const int total_defNum = (int)self_struct->m_defNumSum_ptr[batch_size];
		const int cols = total_defNum;
		//for pointer increment
		const int cross_channel_stride = cols * kernel_dim;

		//init vals to zero
		memset(col_buff, 0, sizeof(Dtype) * cols * rows);


		Dtype *col_ptr = col_buff;

		const unsigned char *offset = self_struct->m_offset_ptr;
		const PACKED_POSITION *position_tags = self_struct->m_posTag_ptr;
		const Dtype *cur_valid_data = valid_data;
		const int *hash_index_table = self_struct->m_hashIdxTable_ptr;

		int counter = 0;	//for debug
		for (int b = 0; b < batch_size; b++)
		{
			const int m_bar = (int)self_struct->m_mBar_ptr[b];
			const int r_bar = (int)self_struct->m_rBar_ptr[b];

			const int m = m_bar * m_bar * m_bar;
			const int r = r_bar * r_bar * r_bar;
			const int r2 = r_bar * r_bar;
			const int m2 = m_bar * m_bar;

			const int dense_res = (int)self_struct->m_res_ptr[b];
			const int defNum = (int)self_struct->m_defNum_ptr[b];

			for (int v = 0; v < m; v++)
			{
				//if the hash voxel is undefined, skip
				if (!ishashVoxelDefined(&position_tags[v]))
				{
					continue;
				}
				///////////////////////////////////////////

				//get the real voxel position from the position tag
				int cx, cy, cz;
				xyz_from_pack(position_tags[v], cx, cy, cz);	//get the voxel position mapped to this hash
																///////////////////////////////////////////////

																//loop over neighbors to fill the column
				int min_x = cx - hx;
				int min_y = cy - hy;
				int min_z = cz - hz;
				int mx, my, mz;
				Dtype *cur_row = col_ptr;
				for (int nz = min_z; nz < min_z + kernel_shape[2]; nz++)
				{
					for (int ny = min_y; ny < min_y + kernel_shape[1]; ny++)
					{
						for (int nx = min_x; nx < min_x + kernel_shape[0]; nx++)
						{
							if (nx < 0 || ny < 0 || nz < 0 || nx >= dense_res || ny >= dense_res || nz >= dense_res)
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}
							//hash to get hash position
							Hash(nx, ny, nz, mx, my, mz,
								offset, m_bar, r_bar, r2);
							const int m_idx = NXYZ2I(mx, my, mz, m_bar, m2);

							if (!ishashVoxelDefined(&position_tags[m_idx]))	//the hash voxel is undefined
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}

							int stored_x, stored_y, stored_z;
							xyz_from_pack(position_tags[m_idx], stored_x, stored_y, stored_z);
							if (nx != stored_x || ny != stored_y || nz != stored_z)	//the neighbor is an undefined voxel
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}

							//fill the value at cur_row and corresponding channels
							int valid_idx = hash_index_table[m_idx];
							if (valid_idx < 0 || valid_idx >= defNum)
							{
								LOG(INFO) << "Fatal error: unexpected wrong hash index " << valid_idx<<"at m_idx "<<m_idx;
								exit(0);
							}
							const Dtype *data_ptr = &cur_valid_data[valid_idx];
							Dtype *fill_row_ptr = cur_row;
							for (int c = 0; c < channels; c++)
							{
								*fill_row_ptr = *data_ptr;
								fill_row_ptr += cross_channel_stride;
								data_ptr += total_defNum;	//NOTE: the data are stored in the blob packing all data
							}


							cur_row += cols;
						}
					}
				}

				counter++;
				col_ptr++;
			}

			offset += r * 3;
			position_tags += m;
			cur_valid_data += defNum;
			hash_index_table += m;
		}

		if (counter != cols)
		{
			LOG(INFO) << "Fatal error: counter " << counter << " != total def num " << cols;
			exit(0);
		}

		return 1;
	}


	template<typename Dtype>
	int col2hash_self_cpu(Dtype* valid_data, int channels, const CBatchStructure *self_struct,
		const int kernel_shape[3],	//D, H, W
		const Dtype* col_buff)
	{
		const int batch_size = self_struct->m_batch_size;
		const int hx = (kernel_shape[0] >> 1);
		const int hy = (kernel_shape[1] >> 1);
		const int hz = (kernel_shape[2] >> 1);
		//col is reception field: input_channels * KD * KH * KW; row: spatial domain
		const int kernel_dim = kernel_shape[0] * kernel_shape[1] * kernel_shape[2];
		const int rows = kernel_dim * channels;
		const int total_defNum = (int)self_struct->m_defNumSum_ptr[batch_size];
		const int cols = total_defNum;
		//for pointer increment
		const int cross_channel_stride = cols * kernel_dim;

		//init vals to zero
		memset(valid_data, 0, sizeof(Dtype) * channels * total_defNum);


		Dtype *cur_valid_data = valid_data;

		const Dtype *col_ptr = col_buff;
		const unsigned char *offset = self_struct->m_offset_ptr;
		const PACKED_POSITION *position_tags = self_struct->m_posTag_ptr;
		const int *hash_index_table = self_struct->m_hashIdxTable_ptr;

		int counter = 0;	//for debug
		for (int b = 0; b < batch_size; b++)
		{
			const int m_bar = (int)self_struct->m_mBar_ptr[b];
			const int r_bar = (int)self_struct->m_rBar_ptr[b];

			const int m = m_bar * m_bar * m_bar;
			const int r = r_bar * r_bar * r_bar;
			const int r2 = r_bar * r_bar;
			const int m2 = m_bar * m_bar;

			const int dense_res = (int)self_struct->m_res_ptr[b];
			const int defNum = (int)self_struct->m_defNum_ptr[b];

			for (int v = 0; v < m; v++)
			{
				//if the hash voxel is undefined, skip
				if (!ishashVoxelDefined(&position_tags[v]))
				{
					continue;
				}
				///////////////////////////////////////////

				//get the real voxel position from the position tag
				int cx, cy, cz;
				xyz_from_pack(position_tags[v], cx, cy, cz);	//get the voxel position mapped to this hash
																///////////////////////////////////////////////

																//loop over neighbors to fill the column
				int min_x = cx - hx;
				int min_y = cy - hy;
				int min_z = cz - hz;
				int mx, my, mz;
				const Dtype *cur_row = col_ptr;
				for (int nz = min_z; nz < min_z + kernel_shape[2]; nz++)
				{
					for (int ny = min_y; ny < min_y + kernel_shape[1]; ny++)
					{
						for (int nx = min_x; nx < min_x + kernel_shape[0]; nx++)
						{
							if (nx < 0 || ny < 0 || nz < 0 || nx >= dense_res || ny >= dense_res || nz >= dense_res)
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}
							//hash to get hash position
							Hash(nx, ny, nz, mx, my, mz,
								offset, m_bar, r_bar, r2);
							const int m_idx = NXYZ2I(mx, my, mz, m_bar, m2);

							if (!ishashVoxelDefined(&position_tags[m_idx]))	//the hash voxel is undefined
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}

							int stored_x, stored_y, stored_z;
							xyz_from_pack(position_tags[m_idx], stored_x, stored_y, stored_z);
							if (nx != stored_x || ny != stored_y || nz != stored_z)	//the neighbor is an undefined voxel
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}

							//fill the value at cur_row and corresponding channels
							int valid_idx = hash_index_table[m_idx];
							if (valid_idx < 0 || valid_idx >= defNum)
							{
								LOG(INFO) << "Fatal error: unexpected wrong hash index " << valid_idx;
								exit(0);
							}
							Dtype *data_ptr = &cur_valid_data[valid_idx];
							const Dtype *fill_row_ptr = cur_row;
							for (int c = 0; c < channels; c++)
							{
								*data_ptr += *fill_row_ptr;
								fill_row_ptr += cross_channel_stride;
								data_ptr += total_defNum;
							}


							cur_row += cols;
						}
					}
				}

				counter++;
				col_ptr++;
			}

			offset += r * 3;
			position_tags += m;
			cur_valid_data += defNum;;
			hash_index_table += m;
		}

		if (counter != cols)
		{
			LOG(INFO) << "Fatal error: counter " << counter << " != total def num " << cols;
			exit(0);
		}

		return 1;
	}


	/**************************************************************************************/
	/*Hash2Col and Col2Hash with stride>1. Not fully tested, just unit test passed***/
	/**************************************************************************************/
	template<typename Dtype>
	int hash2col_ds_cpu(const Dtype* bt_data, int bt_channels, const CBatchStructure *bt_struct,
		const CBatchStructure *top_struct,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		Dtype* col_buff)
	{
		const int batch_size = bt_struct->m_batch_size;
		//col is reception field: input_channels * KD * KH * KW; row: spatial domain
		const int kernel_dim = kernel_shape[0] * kernel_shape[1] * kernel_shape[2];
		const int rows = kernel_dim * bt_channels;
		const int top_total_defNum = (int)top_struct->m_defNumSum_ptr[batch_size];
		const int cols = top_total_defNum;
		//for pointer increment
		const int cross_channel_stride = cols * kernel_dim;

		//init vals to zero
		memset(col_buff, 0, sizeof(Dtype) * cols * rows);


		Dtype *col_ptr = col_buff;

		const unsigned char *bt_offset = bt_struct->m_offset_ptr;
		const PACKED_POSITION *bt_position_tags = bt_struct->m_posTag_ptr;
		const Dtype *cur_bt_data = bt_data;
		const int *bt_hash_index_table = bt_struct->m_hashIdxTable_ptr;
		const int bt_total_defNum = (int)bt_struct->m_defNumSum_ptr[batch_size];
		
		const PACKED_POSITION *tp_position_tags = top_struct->m_posTag_ptr;

		int counter = 0;	//for debug
		for (int b = 0; b < batch_size; b++)
		{
			const int top_m_bar = (int)top_struct->m_mBar_ptr[b];
			const int top_m = top_m_bar * top_m_bar * top_m_bar;

			const int bt_m_bar = (int)bt_struct->m_mBar_ptr[b];
			const int bt_m = bt_m_bar * bt_m_bar * bt_m_bar;
			const int bt_r_bar = (int)bt_struct->m_rBar_ptr[b];
			const int bt_r = bt_r_bar * bt_r_bar * bt_r_bar;
			const int bt_r2 = bt_r_bar * bt_r_bar;
			const int bt_m2 = bt_m_bar * bt_m_bar;

			const int bt_dense_res = (int)bt_struct->m_res_ptr[b];
			const int bt_defNum = (int)bt_struct->m_defNum_ptr[b];

			for (int v = 0; v < top_m; v++)
			{
				//if the hash voxel is undefined, skip
				if (!ishashVoxelDefined(&tp_position_tags[v]))
				{
					continue;
				}
				///////////////////////////////////////////

				//get the real top voxel position from the position tag
				int tx, ty, tz;
				xyz_from_pack(tp_position_tags[v], tx, ty, tz);	//get the top voxel position mapped to this hash
																///////////////////////////////////////////////

				int min_x = tx * strides[0] - pads[0];
				int min_y = ty * strides[1] - pads[1];
				int min_z = tz * strides[2] - pads[2];
				int mx, my, mz;
				Dtype *cur_row = col_ptr;
				for (int nz = min_z; nz < min_z + kernel_shape[2]; nz++)
				{
					for (int ny = min_y; ny < min_y + kernel_shape[1]; ny++)
					{
						for (int nx = min_x; nx < min_x + kernel_shape[0]; nx++)
						{
							if (nx < 0 || ny < 0 || nz < 0 || nx >= bt_dense_res || ny >= bt_dense_res || nz >= bt_dense_res)
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}
							//hash to get hash position
							Hash(nx, ny, nz, mx, my, mz,
								bt_offset, bt_m_bar, bt_r_bar, bt_r2);
							const int bt_m_idx = NXYZ2I(mx, my, mz, bt_m_bar, bt_m2);

							if (!ishashVoxelDefined(&bt_position_tags[bt_m_idx]))	//the hash voxel is undefined
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}

							int stored_x, stored_y, stored_z;
							xyz_from_pack(bt_position_tags[bt_m_idx], stored_x, stored_y, stored_z);
							if (nx != stored_x || ny != stored_y || nz != stored_z)	//the neighbor is an undefined voxel
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}

							//fill the value at cur_row and corresponding channels
							int valid_idx = bt_hash_index_table[bt_m_idx];
							if (valid_idx < 0 || valid_idx >= bt_defNum)
							{
								LOG(INFO) << "Fatal error: unexpected wrong hash index " << valid_idx << "at m_idx " << bt_m_idx;
								exit(0);
							}
							const Dtype *data_ptr = &cur_bt_data[valid_idx];
							Dtype *fill_row_ptr = cur_row;
							for (int c = 0; c < bt_channels; c++)
							{
								*fill_row_ptr = *data_ptr;
								fill_row_ptr += cross_channel_stride;
								data_ptr += bt_total_defNum;	//NOTE: the data are stored in the blob packing all data
							}


							cur_row += cols;
						}
					}
				}

				counter++;
				col_ptr++;
			}

			bt_offset += bt_r * 3;
			bt_position_tags += bt_m;
			cur_bt_data += bt_defNum;
			bt_hash_index_table += bt_m;

			tp_position_tags += top_m;
		}

		if (counter != cols)
		{
			LOG(INFO) << "Fatal error: counter " << counter << " != total def num " << cols;
			exit(0);
		}

		return 1;
	}


	template<typename Dtype>
	int col2hash_ds_cpu(Dtype* bt_data, int bt_channels, const CBatchStructure *bt_struct,
		const CBatchStructure *top_struct,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		const Dtype* col_buff)
	{
		const int batch_size = bt_struct->m_batch_size;
		//col is reception field: input_channels * KD * KH * KW; row: spatial domain
		const int kernel_dim = kernel_shape[0] * kernel_shape[1] * kernel_shape[2];
		const int rows = kernel_dim * bt_channels;
		const int top_total_defNum = (int)top_struct->m_defNumSum_ptr[batch_size];
		const int cols = top_total_defNum;
		//for pointer increment
		const int cross_channel_stride = cols * kernel_dim;

		//init vals to zero
		memset(bt_data, 0, sizeof(Dtype) * bt_channels * top_total_defNum);


		Dtype *cur_bt_data = bt_data;

		const Dtype *col_ptr = col_buff;
		const unsigned char *bt_offset = bt_struct->m_offset_ptr;
		const PACKED_POSITION *bt_position_tags = bt_struct->m_posTag_ptr;
		const int *bt_hash_index_table = bt_struct->m_hashIdxTable_ptr;
		const int bt_total_defNum = (int)bt_struct->m_defNumSum_ptr[batch_size];

		const PACKED_POSITION *tp_position_tags = top_struct->m_posTag_ptr;

		int counter = 0;	//for debug
		for (int b = 0; b < batch_size; b++)
		{
			const int top_m_bar = (int)top_struct->m_mBar_ptr[b];
			const int top_m = top_m_bar * top_m_bar * top_m_bar;

			const int bt_m_bar = (int)bt_struct->m_mBar_ptr[b];
			const int bt_m = bt_m_bar * bt_m_bar * bt_m_bar;
			const int bt_r_bar = (int)bt_struct->m_rBar_ptr[b];
			const int bt_r = bt_r_bar * bt_r_bar * bt_r_bar;
			const int bt_r2 = bt_r_bar * bt_r_bar;
			const int bt_m2 = bt_m_bar * bt_m_bar;

			const int bt_dense_res = (int)bt_struct->m_res_ptr[b];
			const int bt_defNum = (int)bt_struct->m_defNum_ptr[b];

			for (int v = 0; v < top_m; v++)
			{
				//if the hash voxel is undefined, skip
				if (!ishashVoxelDefined(&tp_position_tags[v]))
				{
					continue;
				}
				///////////////////////////////////////////
				//get the real top voxel position from the position tag
				
				int tx, ty, tz;
				xyz_from_pack(tp_position_tags[v], tx, ty, tz);	//get the top voxel position mapped to this hash
																///////////////////////////////////////////////

				int min_x = tx * strides[0] - pads[0];
				int min_y = ty * strides[1] - pads[1];
				int min_z = tz * strides[2] - pads[2];
				int mx, my, mz;
				const Dtype *cur_row = col_ptr;
				for (int nz = min_z; nz < min_z + kernel_shape[2]; nz++)
				{
					for (int ny = min_y; ny < min_y + kernel_shape[1]; ny++)
					{
						for (int nx = min_x; nx < min_x + kernel_shape[0]; nx++)
						{
							if (nx < 0 || ny < 0 || nz < 0 || nx >= bt_dense_res || ny >= bt_dense_res || nz >= bt_dense_res)
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}
							//hash to get hash position
							Hash(nx, ny, nz, mx, my, mz,
								bt_offset, bt_m_bar, bt_r_bar, bt_r2);
							const int bt_m_idx = NXYZ2I(mx, my, mz, bt_m_bar, bt_m2);

							if (!ishashVoxelDefined(&bt_position_tags[bt_m_idx]))	//the hash voxel is undefined
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}

							int stored_x, stored_y, stored_z;
							xyz_from_pack(bt_position_tags[bt_m_idx], stored_x, stored_y, stored_z);
							if (nx != stored_x || ny != stored_y || nz != stored_z)	//the neighbor is an undefined voxel
							{
								//just skip, as the values are inited as zeros
								cur_row += cols;
								continue;
							}

							//fill the value at cur_row and corresponding channels
							int bt_valid_idx = bt_hash_index_table[bt_m_idx];
							if (bt_valid_idx < 0 || bt_valid_idx >= bt_defNum)
							{
								LOG(INFO) << "Fatal error: unexpected wrong hash index " << bt_valid_idx;
								exit(0);
							}
							Dtype *data_ptr = &cur_bt_data[bt_valid_idx];
							const Dtype *fill_row_ptr = cur_row;
							for (int c = 0; c < bt_channels; c++)
							{
								*data_ptr += *fill_row_ptr;
								fill_row_ptr += cross_channel_stride;
								data_ptr += bt_total_defNum;
							}
							cur_row += cols;
						}
					}
				}

				counter++;
				col_ptr++;
			}

			bt_offset += bt_r * 3;
			bt_position_tags += bt_m;
			cur_bt_data += bt_defNum;
			bt_hash_index_table += bt_m;

			tp_position_tags += top_m;
		}

		if (counter != cols)
		{
			LOG(INFO) << "Fatal error: counter " << counter << " != total def num " << cols;
			exit(0);
		}

		return 1;
	}
	/***********************************************************************************************************************/

// Explicit instantiation
	template int hash2col_self_cpu<float>(const float* valid_data, int channels, const CBatchStructure *self_struct,
		const int kernel_shape[3],	//D, H, W
		float* col_buff);
	template int hash2col_self_cpu<double>(const double* valid_data, int channels, const CBatchStructure *self_struct,
		const int kernel_shape[3],	//D, H, W
		double* col_buff);


	template int col2hash_self_cpu<float>(float* valid_data, int channels, const CBatchStructure *self_struct,
		const int kernel_shape[3],	//D, H, W
		const float* col_buff);
	template int col2hash_self_cpu<double>(double* valid_data, int channels, const CBatchStructure *self_struct,
		const int kernel_shape[3],	//D, H, W
		const double* col_buff);

	template int hash2col_ds_cpu<float>(const float* bt_data, int bt_channels, const CBatchStructure *bt_struct, const CBatchStructure *top_struct,
		const int kernel_shape[3],	const int stride[3], const int pad[3],
		float* col_buff);
	template int hash2col_ds_cpu<double>(const double* bt_data, int bt_channels, const CBatchStructure *bt_struct, const CBatchStructure *top_struct,
		const int kernel_shape[3],	const int stride[3], const int pad[3],
		double* col_buff);

	template int col2hash_ds_cpu<float>(float* bt_data, int bt_channels, const CBatchStructure *bt_struct, const CBatchStructure *top_struct, 
		const int kernel_shape[3],	const int stride[3], const int pad[3],
		const float* col_buff);
	template int col2hash_ds_cpu<double>(double* bt_data, int bt_channels, const CBatchStructure *bt_struct, const CBatchStructure *top_struct,
		const int kernel_shape[3],	const int stride[3], const int pad[3],
		const double* col_buff);
}  // namespace caffe
