#include <algorithm>

#include "caffe/common.hpp"
#include "caffe/util/hash2col.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/util/gpu_util.cuh"


namespace caffe {
	template <typename Dtype>
	__global__ void hash2col_self_gpu_kernel(const Dtype* valid_data, int channels, int kernel_spatial_dim,
		int total_m, int total_def_num,
		int3 kernel_shape,
		const CBatchStructure self_struct,
		Dtype* col_buff)
	{
		//nThreads = channels * kernel_shape[0] * kernel_shape[1] * kernel_shape[2] * mSum
		CUDA_KERNEL_LOOP(cur_thread, channels * kernel_spatial_dim * total_m)
		{
			int M_idx, K, C;
			I2NXYZ_g(cur_thread, M_idx, K, C, total_m, kernel_spatial_dim * total_m);

			const int local_col_idx = self_struct.m_hashIdxTable_ptr[M_idx];	//valid idx in the hash within the batch
			const VolumeIndexType volIdx = self_struct.m_volIdx_ptr[M_idx];	//the idx of the hash in the batch
			const int dense_res = (int)self_struct.m_res_ptr[volIdx];	//dense resolution
			const int m_bar = (int)self_struct.m_mBar_ptr[volIdx];	//m bar
			const int r_bar = (int)self_struct.m_rBar_ptr[volIdx];	//r bar
			const int mSum = (int)self_struct.m_mSum_ptr[volIdx];	//mSum before this hash
			const int rSum = (int)self_struct.m_rSum_ptr[volIdx];	//rSum before this hash
			const int defNumSum = (int)self_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash

			if (local_col_idx < 0)
			{
				return;	//invalid hash slot. No corresponding column in the buffer
			}

			const int col_idx = defNumSum + local_col_idx;	//column in the col buf
			const int col_buf_idx = NXYZ2I_g(col_idx, K, C, total_def_num, kernel_spatial_dim*total_def_num);	//index in the col buf

			/************get the voxel position mapped to this hash slot***********/
			const int3 center = xyz_from_pack_g(self_struct.m_posTag_ptr[M_idx]);

			/****************compute the neighbor <nx,ny,nz>*************/
			const int min_x = center.x - (kernel_shape.x >> 1);
			const int min_y = center.y - (kernel_shape.y >> 1);
			const int min_z = center.z - (kernel_shape.z >> 1);

			int kx, ky, kz;
			I2NXYZ_g(K, kx, ky, kz, kernel_shape.x, kernel_shape.x * kernel_shape.y);

			const int nx = min_x + kx;
			const int ny = min_y + ky;
			const int nz = min_z + kz;

			/*****************get the data of the neighbor***************/
			
			if (nx < 0 || ny < 0 || nz < 0 || nx >= dense_res || ny >= dense_res || nz >= dense_res)
			{
				col_buff[col_buf_idx] = 0;	//fill with 0
				return;
			}

			//hash to get hash position

			const unsigned char* offset = self_struct.m_offset_ptr + rSum * 3;
			const PACKED_POSITION *posTags = self_struct.m_posTag_ptr + mSum;
			const int *hash_index_table = self_struct.m_hashIdxTable_ptr + mSum;

			//neighbor m
			const int3 mxyz = Hash_g(nx, ny, nz, offset, m_bar, r_bar);
			const int nei_m_idx = NXYZ2I_g(mxyz.x, mxyz.y, mxyz.z, m_bar);

			//neighbor not defined
			const int nei_valid_idx = hash_index_table[nei_m_idx];
			if (nei_valid_idx < 0)
			{
				col_buff[col_buf_idx] = 0;	//fill with 0
				return;	
			}
			//position conflict with stored position
			const int3 stored = xyz_from_pack_g(posTags[nei_m_idx]);
			if (nx != stored.x || ny != stored.y || nz != stored.z)
			{
				col_buff[col_buf_idx] = 0;	//fill with 0
				return;
			}
			
			//fill the col buf
			const int NEI_idx = defNumSum + nei_valid_idx;	//index of neighbor data in batch_valid_data
			col_buff[col_buf_idx] = valid_data[C * total_def_num + NEI_idx];
		}
	}

#if RECORD_NEIGHBOR
	template <typename Dtype>
	__global__ void hash2col_self_gpu_kernel(const Dtype* valid_data, int channels, int kernel_spatial_dim,
		int total_m, int total_def_num,
		int3 kernel_shape,
		const CBatchStructure self_struct,
		Dtype* col_buff, int *nei_indices)
	{
		//nThreads = channels * kernel_shape[0] * kernel_shape[1] * kernel_shape[2] * mSum
		CUDA_KERNEL_LOOP(cur_thread, channels * kernel_spatial_dim * total_m)
		{
			int M_idx, K, C;
			I2NXYZ_g(cur_thread, M_idx, K, C, total_m, kernel_spatial_dim * total_m);

			const int local_col_idx = self_struct.m_hashIdxTable_ptr[M_idx];	//valid idx in the hash within the batch
			const VolumeIndexType volIdx = self_struct.m_volIdx_ptr[M_idx];	//the idx of the hash in the batch
			const int dense_res = (int)self_struct.m_res_ptr[volIdx];	//dense resolution
			const int m_bar = (int)self_struct.m_mBar_ptr[volIdx];	//m bar
			const int r_bar = (int)self_struct.m_rBar_ptr[volIdx];	//r bar
			const int mSum = (int)self_struct.m_mSum_ptr[volIdx];	//mSum before this hash
			const int rSum = (int)self_struct.m_rSum_ptr[volIdx];	//rSum before this hash
			const int defNumSum = (int)self_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash

			if (local_col_idx < 0)
			{
				return;	//invalid hash slot. No corresponding column in the buffer
			}

			const int col_idx = defNumSum + local_col_idx;	//column in the col buf
			const int col_buf_idx = NXYZ2I_g(col_idx, K, C, total_def_num, kernel_spatial_dim*total_def_num);	//index in the col buf

																												/************get the voxel position mapped to this hash slot***********/
			const int3 center = xyz_from_pack_g(self_struct.m_posTag_ptr[M_idx]);

			/****************compute the neighbor <nx,ny,nz>*************/
			const int min_x = center.x - (kernel_shape.x >> 1);
			const int min_y = center.y - (kernel_shape.y >> 1);
			const int min_z = center.z - (kernel_shape.z >> 1);

			int kx, ky, kz;
			I2NXYZ_g(K, kx, ky, kz, kernel_shape.x, kernel_shape.x * kernel_shape.y);

			const int nx = min_x + kx;
			const int ny = min_y + ky;
			const int nz = min_z + kz;

			/*****************get the data of the neighbor***************/

			if (nx < 0 || ny < 0 || nz < 0 || nx >= dense_res || ny >= dense_res || nz >= dense_res)
			{
				col_buff[col_buf_idx] = 0;	//fill with 0
				return;
			}

			//hash to get hash position

			const unsigned char* offset = self_struct.m_offset_ptr + rSum * 3;
			const PACKED_POSITION *posTags = self_struct.m_posTag_ptr + mSum;
			const int *hash_index_table = self_struct.m_hashIdxTable_ptr + mSum;

			//neighbor m
			const int3 mxyz = Hash_g(nx, ny, nz, offset, m_bar, r_bar);
			const int nei_m_idx = NXYZ2I_g(mxyz.x, mxyz.y, mxyz.z, m_bar);

			//neighbor not defined
			const int nei_valid_idx = hash_index_table[nei_m_idx];
			if (nei_valid_idx < 0)
			{
				col_buff[col_buf_idx] = 0;	//fill with 0
				return;
			}
			//position conflict with stored position
			const int3 stored = xyz_from_pack_g(posTags[nei_m_idx]);
			if (nx != stored.x || ny != stored.y || nz != stored.z)
			{
				col_buff[col_buf_idx] = 0;	//fill with 0
				return;
			}

			//fill the col buf
			const int NEI_idx = defNumSum + nei_valid_idx;	//index of neighbor data in batch_valid_data
			col_buff[col_buf_idx] = valid_data[C * total_def_num + NEI_idx];

			//record the neighbor indices
			nei_indices[K*total_def_num + col_idx] = NEI_idx;
		}
	}
#endif

	template<typename Dtype>
	int hash2col_self_gpu(const Dtype* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		Dtype* col_buff)
	{
		int kernel_spatial_dim = kernel_shape[0] * kernel_shape[1] * kernel_shape[2];

		/*************Now we skip the memset zero, and set zero in the kernel. a little bit faster*********/
		//init vals to zero
		int rows = channels * kernel_spatial_dim;
		/*		int cols = total_def_num;
		caffe_gpu_set(rows*cols, Dtype(0), col_buff);	*/

		int nThreads = rows * total_m;

		hash2col_self_gpu_kernel<Dtype> << <CAFFE_GET_BLOCKS(nThreads), CAFFE_CUDA_NUM_THREADS >> >(
			valid_data, channels, kernel_spatial_dim,
			total_m, total_def_num,
			make_int3(kernel_shape[0], kernel_shape[1], kernel_shape[2]),
			*self_struct_GPU, col_buff);
		return 1;
	}

#if RECORD_NEIGHBOR
	template<typename Dtype>
	int hash2col_self_gpu(const Dtype* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		Dtype* col_buff, int *nei_indices)
	{
		int kernel_spatial_dim = kernel_shape[0] * kernel_shape[1] * kernel_shape[2];
		
		/*************Now we skip the memset zero, and set zero in the kernel. a little bit faster*********/
		//init vals to zero
		int rows = channels * kernel_spatial_dim;
/*		int cols = total_def_num;
		caffe_gpu_set(rows*cols, Dtype(0), col_buff);	*/	

		int nThreads =  rows * total_m;

		hash2col_self_gpu_kernel<Dtype> << <CAFFE_GET_BLOCKS(nThreads), CAFFE_CUDA_NUM_THREADS >> >(
			valid_data, channels, kernel_spatial_dim,
			total_m, total_def_num,
			make_int3(kernel_shape[0], kernel_shape[1], kernel_shape[2]),
			*self_struct_GPU, col_buff, nei_indices);
		return 1;
	}
#endif

	template <typename Dtype>
	__global__ void col2hash_self_gpu_kernel(Dtype* valid_data, int channels, int kernel_spatial_dim,
		int total_m, int total_def_num,
		int3 kernel_shape,
		const CBatchStructure self_struct,
		const Dtype* col_buff)
	{
		//nThreads = channels * kernel_shape[0] * kernel_shape[1] * kernel_shape[2] * mSum
		CUDA_KERNEL_LOOP(cur_thread, channels * kernel_spatial_dim * total_m)
		{
			int M_idx, K, C;
			I2NXYZ_g(cur_thread, M_idx, K, C, total_m, kernel_spatial_dim * total_m);

			const int local_col_idx = self_struct.m_hashIdxTable_ptr[M_idx];	//valid idx in the hash within the batch
			const VolumeIndexType volIdx = self_struct.m_volIdx_ptr[M_idx];	//the idx of the hash in the batch
			const int dense_res = (int)self_struct.m_res_ptr[volIdx];	//dense resolution
			const int m_bar = (int)self_struct.m_mBar_ptr[volIdx];	//m bar
			const int r_bar = (int)self_struct.m_rBar_ptr[volIdx];	//r bar
			const int mSum = (int)self_struct.m_mSum_ptr[volIdx];	//mSum before this hash
			const int rSum = (int)self_struct.m_rSum_ptr[volIdx];	//rSum before this hash
			const int defNumSum = (int)self_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash

			if (local_col_idx < 0)
			{
				return;	//invalid hash slot. No corresponding column in the buffer
			}

			const int col_idx = defNumSum + local_col_idx;	//column in the col buf
			const int col_buf_idx = NXYZ2I_g(col_idx, K, C, total_def_num, kernel_spatial_dim*total_def_num);	//index in the col buf

			/************get the voxel position mapped to this hash slot***********/
			const int3 center = xyz_from_pack_g(self_struct.m_posTag_ptr[M_idx]);

			/****************compute the neighbor <nx,ny,nz>*************/
			const int min_x = center.x - (kernel_shape.x >> 1);
			const int min_y = center.y - (kernel_shape.y >> 1);
			const int min_z = center.z - (kernel_shape.z >> 1);

			int kx, ky, kz;
			I2NXYZ_g(K, kx, ky, kz, kernel_shape.x, kernel_shape.x * kernel_shape.y);

			const int nx = min_x + kx;
			const int ny = min_y + ky;
			const int nz = min_z + kz;

			/*****************get the data of the neighbor***************/
			if (nx < 0 || ny < 0 || nz < 0 || nx >= dense_res || ny >= dense_res || nz >= dense_res)
			{
				return;
			}

			//hash to get hash position
			const unsigned char* offset = self_struct.m_offset_ptr + rSum * 3;
			const PACKED_POSITION *posTags = self_struct.m_posTag_ptr + mSum;
			const int *hash_index_table = self_struct.m_hashIdxTable_ptr + mSum;

			//neighbor m
			const int3 mxyz = Hash_g(nx, ny, nz, offset, m_bar, r_bar);
			const int nei_m_idx = NXYZ2I_g(mxyz.x, mxyz.y, mxyz.z, m_bar);

			//neighbor not defined
			const int nei_valid_idx = hash_index_table[nei_m_idx];
			if (nei_valid_idx < 0)
			{
				return;
			}
			//position conflict with stored position
			const int3 stored = xyz_from_pack_g(posTags[nei_m_idx]);
			if (nx != stored.x || ny != stored.y || nz != stored.z)
			{
				return;
			}

			//fill the col buf
			const int NEI_idx = defNumSum + nei_valid_idx;	//index of neighbor data in batch_valid_data


			caffe_gpu_atomic_add(col_buff[col_buf_idx], valid_data + C * total_def_num + NEI_idx);
		}
	}

	template <typename Dtype>
	__global__ void col2hash_self_gpu_kernel(Dtype* valid_data, int channels, int kernel_spatial_dim,
		int total_m, int total_def_num,
		int3 kernel_shape,
		const CBatchStructure self_struct,
		const Dtype* col_buff, const int *nei_indices)
	{
		//nThreads = channels * kernel_shape[0] * kernel_shape[1] * kernel_shape[2] * mSum
		CUDA_KERNEL_LOOP(cur_thread, channels * kernel_spatial_dim * total_m)
		{
			int M_idx, K, C;
			I2NXYZ_g(cur_thread, M_idx, K, C, total_m, kernel_spatial_dim * total_m);

			const int local_col_idx = self_struct.m_hashIdxTable_ptr[M_idx];	//valid idx in the hash within the batch
			const VolumeIndexType volIdx = self_struct.m_volIdx_ptr[M_idx];	//the idx of the hash in the batch
			const int defNumSum = (int)self_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash
			if (local_col_idx < 0)
			{
				return;	//invalid hash slot. No corresponding column in the buffer
			}

			const int col_idx = defNumSum + local_col_idx;	//column in the col buf

			const int NEI_idx = nei_indices[K*total_def_num + col_idx];	//global nei idx

			const int col_buf_idx = NXYZ2I_g(col_idx, K, C, total_def_num, kernel_spatial_dim*total_def_num);	//index in the col buf

			caffe_gpu_atomic_add(col_buff[col_buf_idx], valid_data + C * total_def_num + NEI_idx);
		}
	}

	template<typename Dtype>
	int col2hash_self_gpu(Dtype* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const Dtype* col_buff)
	{
		int kernel_spatial_dim = kernel_shape[0] * kernel_shape[1] * kernel_shape[2];

		int rows = channels * kernel_spatial_dim;
		//int cols = total_def_num;
		caffe_gpu_set(channels*total_def_num, Dtype(0), valid_data);	//NOTE: here we need to set the bottom dif to zero, as in the kernel there is accumulation

		int nThreads = rows * total_m;
		col2hash_self_gpu_kernel<Dtype> << <CAFFE_GET_BLOCKS(nThreads), CAFFE_CUDA_NUM_THREADS >> >(
			valid_data, channels, kernel_spatial_dim,
			total_m, total_def_num,
			make_int3(kernel_shape[0], kernel_shape[1], kernel_shape[2]),
			*self_struct_GPU, col_buff);
		return 1;
	}

	template<typename Dtype>
	int col2hash_self_gpu(Dtype* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const Dtype* col_buff, const int *nei_indices)
	{
		int kernel_spatial_dim = kernel_shape[0] * kernel_shape[1] * kernel_shape[2];

		int rows = channels * kernel_spatial_dim;
		//int cols = total_def_num;
		caffe_gpu_set(channels*total_def_num, Dtype(0), valid_data);	//NOTE: here we need to set the bottom dif to zero, as in the kernel there is accumulation

		int nThreads = rows * total_m;
		col2hash_self_gpu_kernel<Dtype> << <CAFFE_GET_BLOCKS(nThreads), CAFFE_CUDA_NUM_THREADS >> >(
			valid_data, channels, kernel_spatial_dim,
			total_m, total_def_num,
			make_int3(kernel_shape[0], kernel_shape[1], kernel_shape[2]),
			*self_struct_GPU, col_buff, nei_indices);
		return 1;
	}


	/**************************************************************************************/
	/*Hash2Col and Col2Hash with stride>1. Not fully tested, just unit test passed***/
	/**************************************************************************************/
	template <typename Dtype>
	__global__ void hash2col_ds_gpu_kernel(const Dtype* bt_data, int bt_channels, int kernel_spatial_dim,
		int tp_total_m, int tp_total_def_num, int bt_total_def_num,
		int3 kernel_shape, int3 strides, int3 pads,
		const CBatchStructure bt_struct,
		const CBatchStructure tp_struct,
		Dtype* col_buff)
	{
		//nThreads = channels * kernel_shape[0] * kernel_shape[1] * kernel_shape[2] * mSum
		CUDA_KERNEL_LOOP(cur_thread, bt_channels * kernel_spatial_dim * tp_total_m)
		{
			int tp_M_idx, K, C;
			I2NXYZ_g(cur_thread, tp_M_idx, K, C, tp_total_m, kernel_spatial_dim * tp_total_m);

			const int tp_local_col_idx = tp_struct.m_hashIdxTable_ptr[tp_M_idx];	//valid idx in the hash within the batch
			const VolumeIndexType volIdx = tp_struct.m_volIdx_ptr[tp_M_idx];	//the idx of the hash in the batch
			const int tp_defNumSum = (int)tp_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash

			const int bt_dense_res = (int)bt_struct.m_res_ptr[volIdx];	//dense resolution
			const int bt_m_bar = (int)bt_struct.m_mBar_ptr[volIdx];	//m bar
			const int bt_r_bar = (int)bt_struct.m_rBar_ptr[volIdx];	//r bar
			const int bt_mSum = (int)bt_struct.m_mSum_ptr[volIdx];	//mSum before this hash
			const int bt_rSum = (int)bt_struct.m_rSum_ptr[volIdx];	//rSum before this hash
			const int bt_defNumSum = (int)bt_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash

			if (tp_local_col_idx < 0)
			{
				return;	//invalid hash slot. No corresponding column in the buffer
			}

			const int col_idx = tp_defNumSum + tp_local_col_idx;	//column in the col buf
			const int col_buf_idx = NXYZ2I_g(col_idx, K, C, tp_total_def_num, kernel_spatial_dim*tp_total_def_num);	//index in the col buf

																												/************get the voxel position mapped to this hash slot***********/
			const int3 out_voxel_pos = xyz_from_pack_g(tp_struct.m_posTag_ptr[tp_M_idx]);

			/****************compute the neighbor <nx,ny,nz>*************/
			const int min_x = out_voxel_pos.x * strides.x - pads.x;
			const int min_y = out_voxel_pos.y * strides.y - pads.y;
			const int min_z = out_voxel_pos.z * strides.z - pads.z;

			int kx, ky, kz;
			I2NXYZ_g(K, kx, ky, kz, kernel_shape.x, kernel_shape.x * kernel_shape.y);

			const int nx = min_x + kx;
			const int ny = min_y + ky;
			const int nz = min_z + kz;

			/*****************get the data of the input receptive field***************/
			if (nx < 0 || ny < 0 || nz < 0 || nx >= bt_dense_res || ny >= bt_dense_res || nz >= bt_dense_res)
			{
				col_buff[col_buf_idx] = 0;	//fill with 0
				return;
			}

			//hash to get hash position

			const unsigned char* bt_offset = bt_struct.m_offset_ptr + bt_rSum * 3;
			const PACKED_POSITION *bt_posTags = bt_struct.m_posTag_ptr + bt_mSum;
			const int *bt_hash_index_table = bt_struct.m_hashIdxTable_ptr + bt_mSum;

			//bt m
			const int3 mxyz = Hash_g(nx, ny, nz, bt_offset, bt_m_bar, bt_r_bar);
			const int bt_m_idx = NXYZ2I_g(mxyz.x, mxyz.y, mxyz.z, bt_m_bar);

			//bt voxel not defined
			const int bt_valid_idx = bt_hash_index_table[bt_m_idx];
			if (bt_valid_idx < 0)
			{
				col_buff[col_buf_idx] = 0;	//fill with 0
				return;
			}
			//position conflict with stored position
			const int3 stored = xyz_from_pack_g(bt_posTags[bt_m_idx]);
			if (nx != stored.x || ny != stored.y || nz != stored.z)
			{
				col_buff[col_buf_idx] = 0;	//fill with 0
				return;
			}

			//fill the col buf
			const int BT_idx = bt_defNumSum + bt_valid_idx;	//index of bt data in batch_bt_data
			col_buff[col_buf_idx] = bt_data[C * bt_total_def_num + BT_idx];
		}
	}
	template<typename Dtype>
	int hash2col_ds_gpu(const Dtype* bt_data, int bt_channels,
		int tp_total_m, int tp_total_def_num, int bt_total_def_num,
		const CBatchStructure *bt_struct_GPU,
		const CBatchStructure *tp_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		Dtype* col_buff)
	{
		int kernel_spatial_dim = kernel_shape[0] * kernel_shape[1] * kernel_shape[2];

		/*************Now we skip the memset zero, and set zero in the kernel. a little bit faster*********/
		//init vals to zero
		int rows = bt_channels * kernel_spatial_dim;
		/*		int cols = total_def_num;
		caffe_gpu_set(rows*cols, Dtype(0), col_buff);	*/

		int nThreads = rows * tp_total_m;

		hash2col_ds_gpu_kernel<Dtype> << <CAFFE_GET_BLOCKS(nThreads), CAFFE_CUDA_NUM_THREADS >> >(
			bt_data, bt_channels, kernel_spatial_dim,
			tp_total_m, tp_total_def_num, bt_total_def_num,
			make_int3(kernel_shape[0], kernel_shape[1], kernel_shape[2]),
			make_int3(strides[0], strides[1], strides[2]),
			make_int3(pads[0], pads[1], pads[2]),
			*bt_struct_GPU, *tp_struct_GPU, col_buff);
		return 1;
	}


	template <typename Dtype>
	__global__ void col2hash_ds_gpu_kernel(Dtype* bt_data, int bt_channels, int kernel_spatial_dim,
		int tp_total_m, int tp_total_def_num, int bt_total_def_num,
		int3 kernel_shape, int3 strides, int3 pads,
		const CBatchStructure bt_struct,
		const CBatchStructure tp_struct,
		const Dtype* col_buff)
	{
		//nThreads = channels * kernel_shape[0] * kernel_shape[1] * kernel_shape[2] * mSum
		CUDA_KERNEL_LOOP(cur_thread, bt_channels * kernel_spatial_dim * tp_total_m)
		{
			int tp_M_idx, K, C;
			I2NXYZ_g(cur_thread, tp_M_idx, K, C, tp_total_m, kernel_spatial_dim * tp_total_m);

			const int tp_local_col_idx = tp_struct.m_hashIdxTable_ptr[tp_M_idx];	//valid idx in the hash within the batch
			const VolumeIndexType volIdx = tp_struct.m_volIdx_ptr[tp_M_idx];	//the idx of the hash in the batch
			const int tp_defNumSum = (int)tp_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash

			const int bt_dense_res = (int)bt_struct.m_res_ptr[volIdx];	//dense resolution
			const int bt_m_bar = (int)bt_struct.m_mBar_ptr[volIdx];	//m bar
			const int bt_r_bar = (int)bt_struct.m_rBar_ptr[volIdx];	//r bar
			const int bt_mSum = (int)bt_struct.m_mSum_ptr[volIdx];	//mSum before this hash
			const int bt_rSum = (int)bt_struct.m_rSum_ptr[volIdx];	//rSum before this hash
			const int bt_defNumSum = (int)bt_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash

			if (tp_local_col_idx < 0)
			{
				return;	//invalid hash slot. No corresponding column in the buffer
			}

			const int col_idx = tp_defNumSum + tp_local_col_idx;	//column in the col buf
			const int col_buf_idx = NXYZ2I_g(col_idx, K, C, tp_total_def_num, kernel_spatial_dim*tp_total_def_num);	//index in the col buf

																													/************get the voxel position mapped to this hash slot***********/
			const int3 out_voxel_pos = xyz_from_pack_g(tp_struct.m_posTag_ptr[tp_M_idx]);

			/****************compute the neighbor <nx,ny,nz>*************/
			const int min_x = out_voxel_pos.x * strides.x - pads.x;
			const int min_y = out_voxel_pos.y * strides.y - pads.y;
			const int min_z = out_voxel_pos.z * strides.z - pads.z;

			int kx, ky, kz;
			I2NXYZ_g(K, kx, ky, kz, kernel_shape.x, kernel_shape.x * kernel_shape.y);

			const int nx = min_x + kx;
			const int ny = min_y + ky;
			const int nz = min_z + kz;

			/*****************get the data of the input receptive field***************/
			if (nx < 0 || ny < 0 || nz < 0 || nx >= bt_dense_res || ny >= bt_dense_res || nz >= bt_dense_res)
			{
				return;
			}

			//hash to get hash position

			const unsigned char* bt_offset = bt_struct.m_offset_ptr + bt_rSum * 3;
			const PACKED_POSITION *bt_posTags = bt_struct.m_posTag_ptr + bt_mSum;
			const int *bt_hash_index_table = bt_struct.m_hashIdxTable_ptr + bt_mSum;

			//bt m
			const int3 mxyz = Hash_g(nx, ny, nz, bt_offset, bt_m_bar, bt_r_bar);
			const int bt_m_idx = NXYZ2I_g(mxyz.x, mxyz.y, mxyz.z, bt_m_bar);

			//bt voxel not defined
			const int bt_valid_idx = bt_hash_index_table[bt_m_idx];
			if (bt_valid_idx < 0)
			{
				return;
			}
			//position conflict with stored position
			const int3 stored = xyz_from_pack_g(bt_posTags[bt_m_idx]);
			if (nx != stored.x || ny != stored.y || nz != stored.z)
			{
				return;
			}

			//fill the col buf
			const int BT_idx = bt_defNumSum + bt_valid_idx;	//index of bt data in batch_bt_data
			caffe_gpu_atomic_add(col_buff[col_buf_idx], bt_data + C * bt_total_def_num + BT_idx);
		}
	}

	template<typename Dtype>
	int col2hash_ds_gpu(Dtype* bt_data, int bt_channels,
		int tp_total_m, int tp_total_def_num, int bt_total_def_num,
		const CBatchStructure *bt_struct_GPU,
		const CBatchStructure *tp_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		const Dtype* col_buff)
	{
		int kernel_spatial_dim = kernel_shape[0] * kernel_shape[1] * kernel_shape[2];

		int rows = bt_channels * kernel_spatial_dim;
		//int cols = total_def_num;
		caffe_gpu_set(bt_channels*bt_total_def_num, Dtype(0), bt_data);	//NOTE: here we need to set the bottom dif to zero, as in the kernel there is accumulation

		int nThreads = rows * tp_total_m;
		col2hash_ds_gpu_kernel<Dtype> << <CAFFE_GET_BLOCKS(nThreads), CAFFE_CUDA_NUM_THREADS >> >(
			bt_data, bt_channels, kernel_spatial_dim,
			tp_total_m, tp_total_def_num, bt_total_def_num,
			make_int3(kernel_shape[0], kernel_shape[1], kernel_shape[2]),
			make_int3(strides[0], strides[1], strides[2]),
			make_int3(pads[0], pads[1], pads[2]),
			*bt_struct_GPU, *tp_struct_GPU, col_buff);
		return 1;
	}
	/***********************************************************************************************************************/


// Explicit instantiation
	template int hash2col_self_gpu<float>(const float* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		float* col_buff);
	template int hash2col_self_gpu<double>(const double* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		double* col_buff);

	template int col2hash_self_gpu<float>(float* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const float* col_buff);
	template int col2hash_self_gpu<double>(double* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const double* col_buff);

#if RECORD_NEIGHBOR
	template int hash2col_self_gpu<float>(const float* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		float* col_buff, int *nei_indices);
	template int hash2col_self_gpu<double>(const double* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		double* col_buff, int *nei_indices);

	template int col2hash_self_gpu<float>(float* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const float* col_buff, const int *nei_indices);
	template int col2hash_self_gpu<double>(double* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const double* col_buff, const int *nei_indices);
#endif


	template int hash2col_ds_gpu(const float* bt_data, int bt_channels,
		int tp_total_m, int tp_total_def_num, int bt_total_def_num,
		const CBatchStructure *bt_struct_GPU,
		const CBatchStructure *tp_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		float* col_buff);
	template int hash2col_ds_gpu(const double* bt_data, int bt_channels,
		int tp_total_m, int tp_total_def_num, int bt_total_def_num,
		const CBatchStructure *bt_struct_GPU,
		const CBatchStructure *tp_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		double* col_buff);

	template int col2hash_ds_gpu(float* bt_data, int bt_channels,
		int tp_total_m, int tp_total_def_num, int bt_total_def_num,
		const CBatchStructure *bt_struct_GPU,
		const CBatchStructure *tp_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		const float* col_buff);
	template int col2hash_ds_gpu(double* bt_data, int bt_channels,
		int tp_total_m, int tp_total_def_num, int bt_total_def_num,
		const CBatchStructure *bt_struct_GPU,
		const CBatchStructure *tp_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		const double* col_buff);

}  // namespace caffe
